# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

from reahl.tofu import scenario, expected, Fixture
from reahl.tofu.pytestsupport import with_fixtures

from reahl.webdev.tools import Browser

from reahl.component.exceptions import ProgrammerError
from reahl.web.fw import UserInterface
from reahl.web.ui import P, HTML5Page

from reahl.web_dev.fixtures import WebFixture


class UserInterfaceErrorScenarios(Fixture):
    @scenario
    def plug_in_to_nonexistant_name(self):
        self.slot_map = {'name': 'nonexistent'}

    @scenario
    def view_name_not_mapped(self):
        self.slot_map = {}


@with_fixtures(WebFixture, UserInterfaceErrorScenarios)
def test_ui_slots_map_error(web_fixture, user_interface_error_scenarios):

    class SimpleUserInterface(UserInterface):
        def assemble(self):
            root = self.define_view('/', title='View')
            root.set_slot('name', P.factory())

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page)
            self.define_user_interface('/a_ui',  SimpleUserInterface,  user_interface_error_scenarios.slot_map, name='test_ui')

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(ProgrammerError):
        browser.open('/a_ui/')





# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
.. versionadded:: 3.2

Utilities to deal with layout.


"""

import six

from reahl.component.exceptions import arg_checks, IsInstance
from reahl.web.fw import Layout
from reahl.web.ui import Div, HTML5Page, Slot, Header, Footer


class PageLayout(Layout):
    """A PageLayout creates a basic skeleton inside an :class:`reahl.web.ui.HTML5Page`, and optionally
       applies specified :class:`~reahl.web.fw.Layout`\s to parts of this skeleton.

       The skeleton consists of a :class:`~reahl.web.ui.Div` called the `document` of the page, which
       contains three sub-sections inside of it:
       
         - the `.header` -- the page header area where menus and banners go;
         - the `.contents` of the page -- the main area to which the main content will be added; and
         - the `.footer` -- the page footer where links and legal notices go.    

       :keyword document_layout: A :class:`~reahl.web.fw.Layout` that will be applied to `.document`.
       :keyword contents_layout: A :class:`~reahl.web.fw.Layout` that will be applied to `.contents`.
       :keyword header_layout: A :class:`~reahl.web.fw.Layout` that will be applied to `.header`.
       :keyword footer_layout: A :class:`~reahl.web.fw.Layout` that will be applied to `.footer`.

       .. admonition:: Styling
       
          Adds a <div id="doc"> to the <body> of the page, which contains:

           - a <header id="hd">
           - a <div id="contents">
           - a <footer id="ft">

       .. versionadded:: 3.2

    """
    @arg_checks(document_layout=IsInstance(Layout, allow_none=True), contents_layout=IsInstance(Layout, allow_none=True),
                header_layout=IsInstance(Layout, allow_none=True), footer_layout=IsInstance(Layout, allow_none=True))
    def __init__(self, document_layout=None, contents_layout=None, header_layout=None, footer_layout=None):
        super(PageLayout, self).__init__()
        self.header = None    #: The :class:`reahl.web.ui.Header` of the page.
        self.contents = None  #: The :class:`reahl.web.ui.Div` containing the contents.
        self.footer = None    #: The :class:`reahl.web.ui.Footer` of the page.
        self.document = None  #: The :class:`reahl.web.ui.Div` containing the entire page.
        self.header_layout = header_layout   #: A :class:`reahl.web.fw.Layout` to be used for the header of the page.
        self.contents_layout = contents_layout #: A :class:`reahl.web.fw.Layout` to be used for the contents div of the page.
        self.footer_layout = footer_layout   #: A :class:`reahl.web.fw.Layout` to be used for the footer of the page.
        self.document_layout = document_layout #: A :class:`reahl.web.fw.Layout` to be used for the document of the page.

    @arg_checks(widget=IsInstance(HTML5Page))
    def apply_to_widget(self, widget):
        super(PageLayout, self).apply_to_widget(widget)

    def customise_widget(self):
        self.document = self.widget.body.add_child(Div(self.view))
        self.document.set_id('doc')
        if self.document_layout:
            self.document.use_layout(self.document_layout)
        
        self.header = self.document.add_child(Header(self.view))
        self.header.add_child(Slot(self.view, 'header'))
        self.header.set_id('hd')
        if self.header_layout:
            self.header.use_layout(self.header_layout)

        self.contents = self.document.add_child(Div(self.view))
        if self.contents_layout:
            self.contents.use_layout(self.contents_layout)

        self.contents.set_id('bd')
        self.contents.set_attribute('role', 'main')
        
        self.footer = self.document.add_child(Footer(self.view))
        self.footer.add_child(Slot(self.view, 'footer'))
        self.footer.set_id('ft')
        if self.footer_layout:
            self.footer.use_layout(self.footer_layout)




# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The interfaces for persisted classes that are needed by the core Reahl framework. Different
implementations of the framework can be provided by implementing these.
"""
from __future__ import print_function, unicode_literals, absolute_import, division
import six

from abc import ABCMeta, abstractmethod


@six.add_metaclass(ABCMeta)
class UserSessionProtocol(object):
    """A UserSession represents a potentially lengthy interaction of a particular user with the system."""

    @classmethod
    @abstractmethod
    def for_current_session(cls): 
        """Returns a UserSession instance for the current user. If no UserSession is present for the current
           interaction yet this method should create one. If a UserSession does exist for the current interaction,
           this method returns the correct UserSession."""

    @abstractmethod
    def is_active(self): 
        """Answers whether the session is still usable (instead of being expired due to inactivity)."""
        
    @abstractmethod
    def is_secured(self): 
        """Answers whether the interaction is currently done via a secure channel where applicable.
        
           .. versionadded:: 3.1
        """
        
    @abstractmethod
    def set_last_activity_time(self): 
        """Sets a timestamp on the UserSession to indicate when the last activity was detected 
           relating to this interaction. UserSessions typically expire automatically if no activity is
           detected after some time."""

    @abstractmethod
    def get_interface_locale(self): 
        """Answers a string identifying the currently selected locale."""

    @classmethod
    @abstractmethod
    def get_or_create_session(cls): 
        """Returns a session, creating a new one if none can be determined. If one can be determined from
           the browser, that session is returned."""

    @abstractmethod
    def set_session_key(self, response): 
        """Called at the end of a request loop to enable an implementation to save some identifying information
           to the response (such as setting a cookie with the ID of the current session).
        """

    @classmethod
    @abstractmethod
    def initialise_web_session_on(cls, context):
        """TODO"""
        

@six.add_metaclass(ABCMeta)
class SessionDataProtocol(object):
    @classmethod
    @abstractmethod
    def clear_for_form(cls, form):
        """Removes all the user input associated with the given :class:`reahl.web.ui.Form`."""
    
    __hash__ = None
    @abstractmethod
    def __eq__(self, other): 
        """Is required to be implemented."""

    @abstractmethod
    def __neq__(self, other): 
        """Is required to be implemented."""


@six.add_metaclass(ABCMeta)
class UserInputProtocol(SessionDataProtocol):
    """User input, typed as strings on a form is persisted using this class, for the current user's session
       for use in a subsequent request. Used via `web.persisted_userinput_class`.
    """
    @classmethod
    @abstractmethod
    def get_previously_entered_for_form(cls, form, input_name, entered_input_type):
        """Returns the user input associated with the given :class:`reahl.web.ui.Form`, as previously saved using
           `input_name` as name.

           .. versionchanged:: 4.0
              Added entered_input_type
        """

    @classmethod
    @abstractmethod
    def save_input_value_for_form(cls, form, input_name, value, entered_input_type):
        """Persists `value` as the value of the user input associated with the given :class:`reahl.web.ui.Form`,
           using `input_name` as name.

           .. versionchanged:: 4.0
              Added entered_input_type
        """


class PersistedExceptionProtocol(SessionDataProtocol):
    """When a :class:`reahl.component.exceptions.DomainException` happens during Form submission, the 
       exception is persisted using this class, for the current user's session for use in a subsequent 
       request. Used via `web.persisted_exception_class`.
    """
    @classmethod
    @abstractmethod
    def get_exception_for_form(cls, form):
        """Retrieves an exception previously saved for the given :class:`reahl.web.ui.Form`, or None if
           not found."""

    @classmethod
    @abstractmethod
    def get_exception_for_input(cls, form, input_name):
        """Retrieves an exception previously saved for the given :class:`reahl.web.ui.Form` and `input_name`
           or None if not found."""
    
    @classmethod
    @abstractmethod
    def clear_for_all_inputs(cls, form): 
        """Removes all saved Exceptions associated with the given :class:`reahl.web.ui.Form`."""

      
@six.add_metaclass(ABCMeta)
class PersistedFileProtocol(object):
    """When a file is uploaded, file is persisted using this class, for the current user's session 
       for use in a subsequent request. Used via `web.persisted_file_class`.
    """
    @property
    @abstractmethod
    def file_obj(self): 
        """Returns an object with traditional `.read` and `.seek` methods which can be used to
           read the contents of the persisted file.
        """

    @classmethod
    @abstractmethod
    def clear_for_form(cls, form): 
        """Removes all files previously saved for the given :class:`reahl.web.ui.Form`."""

    __hash__ = None
    @abstractmethod
    def __eq__(self, other): 
        """Is required to be implemented."""

    @abstractmethod
    def __neq__(self, other): 
        """Is required to be implemented."""

    @classmethod
    @abstractmethod
    def get_persisted_for_form(cls, form, input_name): 
        """Returns the previously persisted file for the given :class:`reahl.web.ui.Form`,
           using the given `input_name` as name."""

    @classmethod
    @abstractmethod
    def add_persisted_for_form(cls, form, input_name, uploaded_file): 
        """Saves the given `uploaded_file` (a :class:`cgi.FileStorage`) using the given `input_name`
           for the given :class:`reahl.web.ui.Form`.
        """

    @classmethod
    @abstractmethod
    def remove_persisted_for_form(cls, form, input_name, filename): 
        """Removes previously persisted file with name `filename`, as saved for the given 
           :class:`reahl.web.ui.Form` and `input_name`."""

    @classmethod
    @abstractmethod
    def is_uploaded_for_form(cls, form, input_name, filename): 
        """Answers whether a file with name `filename` has previously been saved for the given 
           :class:`reahl.web.ui.Form` and `input_name`."""



# Copyright 2016, 2017, 2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.webdev.tools import XPath

from reahl.component.modelinterface import exposed, BooleanField

from reahl.web.bootstrap.ui import Div, P
from reahl.web.bootstrap.forms import Form, FormLayout, CheckboxInput
from reahl.web.bootstrap.popups import PopupA, CheckCheckboxScript

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class PopupAFixture(Fixture):

    # (note that this xpath ensures that the p is the ONLY content of the dialog)
    poppedup_contents = "//div[@class='modal-body' and count(*)=1]/p[@id='contents']"

    def is_popped_up(self):
        return self.web_fixture.driver_browser.is_visible(self.poppedup_contents)


@with_fixtures(WebFixture, PopupAFixture)
def test_default_behaviour(web_fixture, popup_a_fixture):
    """If you click on the A, a popupwindow opens with its contents the specified
       element on the target page."""

    class PopupTestPanel(Div):
        def __init__(self, view):
            super(PopupTestPanel, self).__init__(view)
            self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
            popup_contents = self.add_child(P(view, text='this is the content of the popup'))
            popup_contents.set_id('contents')

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=PopupTestPanel.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    # The A is rendered correctly
    assert browser.is_element_present("//a[@title='Home page' and text()='Home page' and @href='/']")

    # subsequent behaviour
    browser.click(XPath.link_with_text('Home page'))
    browser.wait_for(popup_a_fixture.is_popped_up)

    #check some bootstrap attributes
    dialog_xpath = "//div[@class='modal fade show' and @tabindex='-1' and @role='dialog']/div[@class='modal-dialog']/div[@class='modal-content']"
    assert browser.is_element_present(dialog_xpath)

    browser.click(XPath.button_labelled('Close'))
    browser.wait_for_not(popup_a_fixture.is_popped_up)


@with_fixtures(WebFixture, PopupAFixture)
def test_customising_dialog_buttons(web_fixture, popup_a_fixture):
    """The buttons of the dialog can be customised."""

    class PopupTestPanel(Div):
        def __init__(self, view):
            super(PopupTestPanel, self).__init__(view)
            popup_a = self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
            popup_a.add_js_button('Butt1')
            popup_a.add_js_button('Butt2')
            popup_contents = self.add_child(P(view, text='this is the content of the popup'))
            popup_contents.set_id('contents')

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=PopupTestPanel.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    button1_xpath = XPath.button_labelled('Butt1')
    button2_xpath = XPath.button_labelled('Butt2')

    browser.open('/')

    browser.click(XPath.link_with_text('Home page'))
    browser.wait_for(popup_a_fixture.is_popped_up)

    assert browser.is_element_present(button1_xpath)
    assert browser.is_element_present(button2_xpath)


@with_fixtures(WebFixture, PopupAFixture)
def test_workings_of_check_checkbox_button(web_fixture, popup_a_fixture):
    """A CheckCheckBoxButton checks the checkbox on the original page when clicked."""

    class PopupTestPanel(Div):
        @exposed
        def fields(self, fields):
            fields.field = BooleanField(label='a checkbox')

        def __init__(self, view):
            super(PopupTestPanel, self).__init__(view)
            popup_a = self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
            popup_contents = self.add_child(P(view, text='this is the content of the popup'))
            popup_contents.set_id('contents')
            form = self.add_child(Form(view, 'aform')).use_layout(FormLayout())
            checkbox = form.layout.add_input(CheckboxInput(form, self.fields.field))

            popup_a.add_js_button('Checkit', CheckCheckboxScript(checkbox))

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=PopupTestPanel.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.click(XPath.link_with_text('Home page'))
    browser.wait_for(popup_a_fixture.is_popped_up)

    browser.click(XPath.button_labelled('Checkit'))
    browser.wait_for_not(popup_a_fixture.is_popped_up)

    assert browser.is_checked(XPath.input_labelled('a checkbox'))


@with_fixtures(WebFixture, PopupAFixture)
def test_centering_dialog_vertically(web_fixture, popup_a_fixture):
    """The dialog can be centered vertically."""

    class PopupTestPanel(Div):
        def __init__(self, view):
            super(PopupTestPanel, self).__init__(view)
            self.add_child(PopupA(view, view.as_bookmark(), '#contents', center_vertically=True))
            popup_contents = self.add_child(P(view, text='this is the content of the popup'))
            popup_contents.set_id('contents')

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=PopupTestPanel.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.click(XPath.link_with_text('Home page'))
    browser.wait_for(popup_a_fixture.is_popped_up)

    dialog_xpath = "//div[@class='modal fade show' and @tabindex='-1' and @role='dialog']/div[@class='modal-dialog modal-dialog-centered']/div[@class='modal-content']"
    assert browser.is_element_present(dialog_xpath)



# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
.. versionadded:: 3.2

Sometimes you need to display a long list of items. Displaying such a list on a single
page is not a good idea, because the page will take forever to load.

This module provides a few classes you can use to build a single View that displays
only one "pageful" of the list. You can then also include a :class:`PageMenu` -- a
menu on which a user can choose to navigate to another section (or page) of the list.



"""

from __future__ import print_function, unicode_literals, absolute_import, division
import six

from functools import partial
from abc import ABCMeta, abstractproperty, abstractmethod

from reahl.component.i18n import Catalogue
from reahl.component.modelinterface import exposed, IntegerField
from reahl.component.decorators import memoized
from reahl.web.fw import Bookmark
from reahl.web.ui import AccessRightAttributes, ActiveStateAttributes, HTMLWidget
from reahl.web.bootstrap.ui import A, Span, Div
from reahl.web.bootstrap.navs import Menu

_ = Catalogue('reahl-web')


@six.add_metaclass(ABCMeta)
class PageIndexProtocol(object):
    @abstractproperty
    def pages_in_range(self): pass
    @abstractproperty
    def current_page(self): pass
    @abstractproperty
    def start_page(self): pass
    @abstractproperty
    def end_page(self): pass
    @abstractproperty
    def previous_page(self): pass
    @abstractproperty
    def next_page(self): pass
    @abstractproperty
    def last_page(self): pass
    @abstractproperty
    def has_next_page(self): pass
    @abstractproperty
    def has_previous_page(self): pass


class PagedPanel(Div):
    """A :class:`Div` whose contents change, depending on the page selected by a user from a :class:`PageMenu`.
       A programmer should subclass from PagedPanel, supplying an `__init__` method which populates the PagedPanel
       with appropriate contents, based on its `.current_contents`.

       .. admonition:: Styling

          Represented in HTML by an <div> element.

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param page_index: The :class:`PageIndex` to use to supply contents to the pages displayed by this PagedPanel.
       :param css_id: (See :class:`reahl.web.ui.HTMLElement`)
    """
    def __init__(self, view, page_index, css_id):
        self.page_index = page_index
        super(PagedPanel, self).__init__(view, css_id=css_id)
        self.enable_refresh()

    def get_bookmark(self, description=None, page_number=1):
        return Bookmark.for_widget(description or '%s' % page_number,
                                   query_arguments={'current_page_number': page_number}).on_view(self.view)

    @exposed
    def query_fields(self, fields):
        fields.current_page_number = self.page_index.fields.current_page_number

    @property
    def current_contents(self):
        """The list of items that should be displayed for the current page."""
        return self.page_index.current_page.contents


class Page(object):
    def __init__(self, number, description, contents_getter):
        self.number = number
        self.description = description
        self.contents_getter = contents_getter

    @property
    def contents(self):
        return self.contents_getter()


class PageIndex(PageIndexProtocol):
    """An object responsible for breaking a long list of items up into shorter lists for display. Each such
       shorter list is referred to as a page. Different ways of breaking long lists into smaller lists
       are provided by subclasses.

       :param current_page_number: The initial page shown to users.
       :param start_page_number: The first page that should be listed in the current :class:`PageMenu`.
       :param max_page_links: The number of pages to be shown in the current :class:`PageMenu`.
    """
    def __init__(self, current_page_number=1, start_page_number=1, max_page_links=5):
        super(PageIndex, self).__init__()
        self.current_page_number = current_page_number
        self.start_page_number = start_page_number
        self.max_page_links = max_page_links

    @exposed
    def fields(self, fields):
        fields.current_page_number = IntegerField(required=False, default=1)
        fields.start_page_number =   IntegerField(required=False, default=1)

    @abstractmethod
    def get_contents_for_page(self, page_number):
        """Override this method in subclasses to obtain the correct list of items for the given `page_number`."""

    @abstractproperty
    def total_number_of_pages(self):
        """Override this @property in subclasses to state what the total number of pages is."""

    def get_description_for_page(self, page_number):
        return six.text_type(page_number)

    @property
    @memoized
    def current_page(self):
        return self.get_page_number(self.current_page_number)

    def get_page_number(self, page_number):
        return Page(page_number, self.get_description_for_page(page_number), partial(self.get_contents_for_page, page_number))

    @property
    @memoized
    def pages_in_range(self):
        return [self.get_page_number(page_number)
                for page_number in list(range(self.start_page_number, self.end_page.number+1))]

    @property
    @memoized
    def start_page(self):
        return self.get_page_number(self.start_page_number)

    @property
    @memoized
    def end_page(self):
        page_number = self.start_page_number+min(self.max_page_links-1, self.total_number_of_pages-self.start_page_number)
        return self.get_page_number(page_number)

    @property
    @memoized
    def previous_page(self):
        page_number = max(1, self.start_page_number - self.max_page_links)
        return self.get_page_number(page_number)

    @property
    @memoized
    def next_page(self):
        page_number = min(self.end_page.number + 1, self.total_number_of_pages)
        return self.get_page_number(page_number)

    @property
    @memoized
    def last_page(self):
        page_number =  max(self.total_number_of_pages - self.max_page_links + 1, 1)
        return self.get_page_number(page_number)

    @property
    @memoized
    def has_next_page(self):
        return self.end_page.number < self.total_number_of_pages

    @property
    @memoized
    def has_previous_page(self):
        return self.start_page_number > 1


class SequentialPageIndex(PageIndex):
    """A PageIndex that breaks a list of items up into smaller lists, by cutting the original list
           into sections that have a maximum number of items per page.

           :param items: The long list of items.
           :keyword items_per_page: The maximum number of items to allow on a page.
           :keyword current_page_number: (See :class:`PageIndex`)
           :keyword start_page_number: (See :class:`PageIndex`)
           :keyword max_page_links: (See :class:`PageIndex`)
        """
    def __init__(self, items, items_per_page=5, current_page_number=1, start_page_number=1, max_page_links=4):
        super(SequentialPageIndex, self).__init__(current_page_number=current_page_number,
                                                  start_page_number=start_page_number,
                                                  max_page_links=max_page_links)
        self.items = items
        self.items_per_page = items_per_page

    def get_contents_for_page(self, page_number):
        range_start = (page_number-1)*self.items_per_page
        range_end = range_start+min(self.items_per_page, len(self.items)-range_start)-1
        return self.items[range_start:range_end+1]

    @property
    @memoized
    def total_number_of_pages(self):
        return ((len(self.items)-1) // self.items_per_page)+1


@six.add_metaclass(ABCMeta)
class AnnualItemOrganiserProtocol(object):
    """Manages a list of items, each of which is seen to be for a particular year.
    """
    @abstractmethod
    def get_years(self):
        """Returns a list of integers, each representing a year which is applicable to at least one item in a list of items."""

    @abstractmethod
    def get_items_for_year(self, year):
        """Returns a list if items to which `year` (an integer) is applicable."""


class AnnualPageIndex(PageIndex):
    """A PageIndex that breaks a list of items up into smaller lists, by arranging all items
       that have the same year on the same page.

       :param annual_item_organiser: An object that implements :class:`AnnualItemOrganiserProtocol`. Its methods
                                     will be called to find the relevent items, or determine what years are applicable.
       :param current_page_number: (See :class:`PageIndex`)
       :param start_page_number: (See :class:`PageIndex`)
       :param max_page_links: (See :class:`PageIndex`)
    """
    def __init__(self, annual_item_organiser, current_page_number=1, start_page_number=1, max_page_links=4):
        super(AnnualPageIndex, self).__init__(current_page_number=current_page_number,
                                              start_page_number=start_page_number,
                                              max_page_links=max_page_links)
        self.annual_item_organiser = annual_item_organiser

    def get_contents_for_page(self, page_number):
        year = self.years[page_number-1]
        return self.annual_item_organiser.get_items_for_year(year)

    def get_description_for_page(self, page_number):
        return six.text_type(self.years[page_number-1])

    @property
    def years(self):
        return self.annual_item_organiser.get_years()

    @property
    @memoized
    def total_number_of_pages(self):
        return len(self.years)


class PageMenu(HTMLWidget):
    """An Menu, which lists the pages of items that can be navigated by a user. If there are
       many pages, only a small subset is shown, with controls allowing the user to browse to
       the wanted page number and choose it.

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param css_id: (See :class:`reahl.web.ui.HTMLElement`)
       :param page_index: The :class:`PageIndex` whose pages are displayed by this PageMenu.
       :param paged_panel: The :class:`PagedPanel` in which the contents of a page is displayed.

    """
    def __init__(self, view, css_id, page_index, paged_panel):
        self.page_index = page_index
        self.paged_panel = paged_panel
        super(PageMenu, self).__init__(view)

        self.menu = self.add_child(Menu(view))
        self.create_items(self.menu)
        self.set_html_representation(self.menu)
        self.set_id(css_id)
        self.append_class('pagination')

        self.enable_refresh(self.query_fields.start_page_number)

    def add_styling_to_menu_item(self, item):
        item.a.append_class('page-link')
        item.html_representation.append_class('page-item')

    def create_items(self, menu):

        self.add_bordering_link_for(menu, '←', 'First', 1,
                                   not self.page_index.has_previous_page)
        self.add_bordering_link_for(menu, '«', 'Prev', self.page_index.previous_page.number, 
                                   not self.page_index.has_previous_page)

        for page in self.page_index.pages_in_range:
            bookmark = self.paged_panel.get_bookmark(page_number=page.number, description=page.description)
            bookmark.query_arguments['start_page_number'] = self.page_index.start_page_number
            link = A.from_bookmark(self.view, bookmark)
            item = menu.add_a(link)
            item.html_representation.add_attribute_source(ActiveStateAttributes(item))
            self.add_styling_to_menu_item(item)
            if self.page_index.current_page_number == page.number:
                item.set_active()

        self.add_bordering_link_for(menu, '»', 'Next', self.page_index.next_page.number, 
                                    not self.page_index.has_next_page)
        self.add_bordering_link_for(menu, '→', 'Last', self.page_index.last_page.number, 
                                    not self.page_index.has_next_page)

    def add_bordering_link_for(self, menu, short_description, long_description, start_page_number, disabled):
        link = A.from_bookmark(self.view, self.get_bookmark(start_page_number=start_page_number, 
                                                            disabled=disabled))

        link.add_child(Span(self.view, text=short_description)).set_attribute('aria-hidden', 'true')
        link.add_child(Span(self.view, text=long_description)).append_class('sr-only')
        link.set_attribute('aria-label', long_description)
        if disabled:
            link.set_attribute('tabindex', '-1')
        link.set_active(not disabled)
        item = menu.add_a(link)
        item.html_representation.add_attribute_source(AccessRightAttributes(link))
        self.add_styling_to_menu_item(item)
        return link
        
    def get_bookmark(self, start_page_number=1, disabled=False):
        bookmark = Bookmark.for_widget(None,
                                       query_arguments={'start_page_number': start_page_number,
                                                        'current_page_number': self.page_index.current_page_number},
                                       write_check=lambda: not disabled).on_view(self.view)
        return bookmark

    @exposed
    def query_fields(self, fields):
        fields.start_page_number = self.page_index.fields.start_page_number
        fields.current_page_number = self.paged_panel.query_fields.current_page_number

    @property
    def jquery_selector(self):
        return '"ul.pagination"'

    def get_js(self, context=None):
        js = ['$(%s).bootstrappagemenu({});' % self.jquery_selector]
        return super(PageMenu, self).get_js(context=context) + js 



# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import scenario, Fixture, uses, expected
from reahl.tofu.pytestsupport import with_fixtures

from reahl.webdev.tools import XPath, Browser
from reahl.webdev.tools import WidgetTester
from reahl.component.modelinterface import exposed, Field, BooleanField, Event, Choice, ChoiceField,\
    MultiChoiceField, IntegerField, Action
from reahl.web.fw import Url
from reahl.web.bootstrap.ui import A, Div, FieldSet
from reahl.web.bootstrap.forms import Button, FormLayout, InlineFormLayout, GridFormLayout, Form, ChoicesLayout,\
    TextInput, CheckboxInput, PrimitiveCheckboxInput, RadioButtonSelectInput, ButtonLayout
from reahl.web.bootstrap.grid import ResponsiveSize

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class FormLayoutScenarios(Fixture):

    def new_widget(self):
        return Div(self.web_fixture.view).use_layout(self.layout)

    @scenario
    def basic_form(self):
        self.layout = FormLayout()
        self.expected_html = '<div></div>'

    @scenario
    def inline_form(self):
        self.layout = InlineFormLayout()
        self.expected_html = '<div class="form-inline"></div>'

    @scenario
    def grid_form(self):
        self.layout = GridFormLayout(ResponsiveSize(xl=4), ResponsiveSize(xl=8))
        self.expected_html = '<div></div>'


@with_fixtures(WebFixture, FormLayoutScenarios)
def test_basic_form_layouts(web_fixture, form_layout_scenarios):
    """There are three basic layouts of forms in bootstrap."""

    tester = WidgetTester(form_layout_scenarios.widget)
    actual = tester.render_html()
    assert actual == form_layout_scenarios.expected_html


@uses(web_fixture=WebFixture)
class FormLayoutFixture(Fixture):
    form_group_xpath = '//form/div[contains(@class, "form-group")]'

    def new_domain_object(self):
        class StubDomainObject(object):
            @exposed
            def fields(self, fields):
                fields.an_attribute = Field(label='Some input', required=True)
        return StubDomainObject()

    def form_contains_form_group(self, browser):
        return browser.get_xpath_count(self.form_group_xpath) == 1

    def get_form_group_children(self, browser, index=0):
        return browser.xpath( '%s[%s]/*' % (self.form_group_xpath, index+1) )

    def get_form_group(self, browser, index=0):
        form_groups = browser.xpath('%s[%s]' % (self.form_group_xpath, index+1))
        return form_groups[index]

    def get_form_group_highlight_marks(self, browser, index=0):
        def is_form_control_element(element):
            return 'class' in element.attrib \
                   and 'form-control' in element.attrib['class']
        [input] = [element for element in self.get_form_group_children(browser, index=index)
                           if is_form_control_element(element)]
        return [mark for mark in input.attrib['class'].split(' ')
                     if mark.startswith('is-')]

    def get_form_group_errors(self, browser, index=0):
        def is_error_element(element):
            return 'class' in element.attrib \
                   and 'invalid-feedback' in element.attrib['class']
        def is_visible(element):
            return not (('style' in element.attrib) and ('display: none' in element.attrib['style']))

        return [element for element in self.get_form_group_children(browser, index=index)
                        if is_error_element(element) and is_visible(element) ]

    def get_label_in_form_group(self, browser, form_group_index=0):
        return browser.xpath('%s[%s]/div/label' % (self.form_group_xpath, form_group_index+1))


@with_fixtures(WebFixture, FormLayoutFixture)
def test_adding_basic_input(web_fixture, form_layout_fixture):
    """Adding an input to a FormLayout, adds it in a bootstrap form-group with Some input."""
    fixture = form_layout_fixture

    class FormWithInputAddedUsingDefaults(Form):
        def __init__(self, view):
            super(FormWithInputAddedUsingDefaults, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute))

    browser = Browser(web_fixture.new_wsgi_app(child_factory=FormWithInputAddedUsingDefaults.factory()))
    browser.open('/')

    assert fixture.form_contains_form_group(browser)

    [label, input_widget] = fixture.get_form_group_children(browser)

    # form-group has a label, correctly set up for bootstrap
    assert label.tag == 'label'
    assert label.attrib['for'] == input_widget.attrib['id']
    assert label.text == 'Some input'

    # form-group has an input, correctly set up for bootstrap
    assert input_widget.tag == 'input'
    assert input_widget.attrib['name'] == 'an_attribute'


@with_fixtures(WebFixture, FormLayoutFixture)
def test_grid_form_layouts(web_fixture, form_layout_fixture):
    """A GridFormLayout adds the Label and Input of each added input in separate columns sized like you specify"""
    fixture = form_layout_fixture

    class FormWithGridFormLayout(Form):
        def __init__(self, view):
            super(FormWithGridFormLayout, self).__init__(view, 'aform')
            self.use_layout(GridFormLayout(ResponsiveSize(lg=4), ResponsiveSize(lg=8)))
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute), help_text='some help')

    browser = Browser(web_fixture.new_wsgi_app(child_factory=FormWithGridFormLayout.factory()))
    browser.open('/')

    [label_column, input_column] = fixture.get_form_group_children(browser)
    assert label_column.tag == 'div'
    assert 'col-lg-4' in label_column.attrib['class'].split(' ')
    assert 'column-label' in label_column.attrib['class'].split(' ')

    [label] = fixture.get_label_in_form_group(browser)
    assert 'col-form-label' in label.attrib['class'].split(' ')

    assert input_column.tag == 'div'
    assert 'col-lg-8' in input_column.attrib['class'].split(' ')
    assert 'column-input' in input_column.attrib['class'].split(' ')

    [input_, help_text] = input_column.getchildren()
    assert input_.tag == 'input'
    assert help_text.tag == 'p'
    assert 'form-text' in help_text.attrib['class'].split(' ')


@with_fixtures(WebFixture, FormLayoutFixture)
def test_inline_form_layouts(web_fixture, form_layout_fixture):
    """An InlineFormLayout adds the Label and Input of each added input next to each other, with some space between them."""
    fixture = form_layout_fixture

    class FormWithInlineFormLayout(Form):
        def __init__(self, view):
            super(FormWithInlineFormLayout, self).__init__(view, 'aform')
            self.use_layout(InlineFormLayout())
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute), help_text='some help')

    browser = Browser(web_fixture.new_wsgi_app(child_factory=FormWithInlineFormLayout.factory()))
    browser.open('/')

    [label, input_, help_text] = fixture.get_form_group_children(browser)

    #check spacing specified between label, input and help_text
    assert label.tag == 'label'
    assert 'mr-2' == label.attrib['class']

    assert input_.tag == 'input'

    assert help_text.tag == 'span'
    assert 'ml-2' in help_text.attrib['class'].split(' ')


@with_fixtures(WebFixture, FormLayoutFixture)
def test_specifying_help_text(web_fixture, form_layout_fixture):
    """You can optionally specify help_text when adding an input."""
    fixture = form_layout_fixture

    class FormWithInputAndHelp(Form):
        def __init__(self, view):
            super(FormWithInputAndHelp, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute), help_text='some help')

    browser = Browser(web_fixture.new_wsgi_app(child_factory=FormWithInputAndHelp.factory()))
    browser.open('/')

    [label, input_widget, help_text] = fixture.get_form_group_children(browser)

    # form-group has help-text
    assert help_text.tag == 'p'
    assert 'text-muted' in help_text.attrib['class']
    assert help_text.text == 'some help'


@with_fixtures(WebFixture, FormLayoutFixture)
def test_omitting_label(web_fixture, form_layout_fixture):
    """The label will be rendered hidden (but available to screen readers) if this is explicity requested."""
    fixture = form_layout_fixture

    class FormWithInputNoLabel(Form):
        def __init__(self, view):
            super(FormWithInputNoLabel, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute), hide_label=True)


    browser = Browser(web_fixture.new_wsgi_app(child_factory=FormWithInputNoLabel.factory()))
    browser.open('/')

    [label, help_text] = fixture.get_form_group_children(browser)
    assert label.tag == 'label'
    assert label.attrib['class'] == 'sr-only'


@with_fixtures(WebFixture, FormLayoutFixture)
def test_adding_checkboxes(web_fixture, form_layout_fixture):
    """CheckboxInputs are added non-inlined, and by default without labels."""

    class DomainObjectWithBoolean(object):
        @exposed
        def fields(self, fields):
            fields.an_attribute = BooleanField(label='Some input', required=True)

    fixture = form_layout_fixture

    fixture.domain_object = DomainObjectWithBoolean()

    class FormWithInputWithCheckbox(Form):
        def __init__(self, view):
            super(FormWithInputWithCheckbox, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(CheckboxInput(self, fixture.domain_object.fields.an_attribute))

    browser = Browser(web_fixture.new_wsgi_app(child_factory=FormWithInputWithCheckbox.factory()))
    browser.open('/')

    assert not any(child.tag == 'label' for child in fixture.get_form_group_children(browser))
    [div] = fixture.get_form_group_children(browser)
    [checkbox] = div.getchildren()
    checkbox_classes = checkbox.attrib['class'].split(' ')
    assert 'custom-control' in checkbox_classes
    assert 'custom-checkbox' in checkbox_classes


class ValidationScenarios(FormLayoutFixture):
    def new_domain_object(self):
        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.an_attribute = Field(label='Some input', required=True)
                fields.another_attribute = Field(label='Another input', required=True)
            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')
        return ModelObject()

    def new_Form(self):
        fixture = self
        class FormWithInput(Form):
            def __init__(self, view):
                super(FormWithInput, self).__init__(view, 'aform')
                self.set_attribute('novalidate', 'novalidate')
                self.use_layout(FormLayout())
                self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute))
                self.layout.add_input(TextInput(self, fixture.domain_object.fields.another_attribute))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))
        return FormWithInput

    @scenario
    def with_javascript(self):
        self.web_fixture.reahl_server.set_app(self.web_fixture.new_wsgi_app(child_factory=self.Form.factory(), enable_js=True))
        self.browser = self.web_fixture.driver_browser

    @scenario
    def without_javascript(self):
        self.browser = Browser(self.web_fixture.new_wsgi_app(child_factory=self.Form.factory()))


@with_fixtures(ValidationScenarios)
def test_input_validation_cues(validation_scenarios):
    """Visible cues are inserted to indicate the current validation state
       and possible validation error messages to a user. """
    fixture = validation_scenarios

    browser = fixture.browser
    browser.open('/')

    assert not fixture.get_form_group_highlight_marks(browser, index=0)
    assert not fixture.get_form_group_errors(browser, index=0)

    browser.type(XPath.input_labelled('Some input'), '')
    browser.click(XPath.button_labelled('Submit'))

    assert ['is-invalid'] == fixture.get_form_group_highlight_marks(browser, index=0)
    [error] = fixture.get_form_group_errors(browser, index=0)
    assert error.text == 'Some input is required'

    browser.type(XPath.input_labelled('Some input'), 'valid value')
    browser.click(XPath.button_labelled('Submit'))

    assert ['is-valid'] == fixture.get_form_group_highlight_marks(browser, index=0)
    assert not fixture.get_form_group_errors(browser, index=0)

    browser.type(XPath.input_labelled('Another input'), 'valid value')
    browser.click(XPath.button_labelled('Submit'))

    assert not fixture.get_form_group_highlight_marks(browser, index=0)
    assert not fixture.get_form_group_errors(browser, index=0)


@with_fixtures(WebFixture, ValidationScenarios.with_javascript)
def test_input_validation_cues_javascript_interaction(web_fixture, javascript_validation_scenario):
    """The visual cues rendered server-side can subsequently be manipulated via javascript."""
    fixture = javascript_validation_scenario

    web_fixture.reahl_server.set_app(web_fixture.new_wsgi_app(child_factory=fixture.Form.factory(), enable_js=False))

    browser = fixture.browser
    browser.open('/')
    browser.type(XPath.input_labelled('Some input'), '')
    browser.click(XPath.button_labelled('Submit'))

    assert ['is-invalid'] == fixture.get_form_group_highlight_marks(browser, index=0)
    [error] = fixture.get_form_group_errors(browser, index=0)
    assert error.text == 'Some input is required'

    web_fixture.reahl_server.set_app(web_fixture.new_wsgi_app(child_factory=fixture.Form.factory(), enable_js=True))
    browser.open('/')

    browser.click(XPath.button_labelled('Submit'))

    assert ['is-invalid'] == fixture.get_form_group_highlight_marks(browser, index=0)
    [error] = fixture.get_form_group_errors(browser, index=0)
    assert error.text == 'Some input is required'

    browser.type(XPath.input_labelled('Some input'), 'valid value')
    browser.press_tab()

    def form_group_is_marked_success(index):
        return ['is-valid'] == fixture.get_form_group_highlight_marks(browser, index=index)
    assert web_fixture.driver_browser.wait_for(form_group_is_marked_success, 0)
    assert not fixture.get_form_group_errors(browser, index=0)


class CheckboxFixture(Fixture):
    def new_field(self):
        return BooleanField(label='Subscribe to newsletter?')

    def new_domain_object(self):
        fixture = self
        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.an_attribute = fixture.field
            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')
        return ModelObject()

    def new_Form(self):
        fixture = self
        class FormWithInput(Form):
            def __init__(self, view):
                super(FormWithInput, self).__init__(view, 'aform')
                self.use_layout(FormLayout())
                self.layout.add_input(CheckboxInput(self, fixture.domain_object.fields.an_attribute))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))
        return FormWithInput


@with_fixtures(WebFixture, CheckboxFixture)
def test_checkbox_basics_with_boolean_field(web_fixture, checkbox_fixture):
    """CheckboxInput can be used to toggle a single yes/no answer, in which case it renders a single checkbox."""

    web_fixture.reahl_server.set_app(web_fixture.new_wsgi_app(child_factory=checkbox_fixture.Form.factory()))
    browser = web_fixture.driver_browser
    browser.open('/')

    checkbox = XPath.input_labelled('Subscribe to newsletter?')
    assert not browser.is_checked(checkbox)
    assert browser.is_element_present(XPath.label_with_text('Subscribe to newsletter?'))
    assert browser.is_element_present('//div/input/following-sibling::label')
    assert browser.get_attribute('//label', 'class') == 'custom-control-label'
    checkbox_classes = browser.get_attribute('//div/input[@type="checkbox"]/..', 'class').split(' ')
    assert 'custom-control' in checkbox_classes
    assert 'custom-checkbox' in checkbox_classes
    assert browser.get_attribute('//div/input[@type="checkbox"]', 'class') == 'custom-control-input'

    browser.check(checkbox)
    browser.click(XPath.button_labelled('Submit'))

    assert browser.is_checked(checkbox)


@with_fixtures(WebFixture, CheckboxFixture)
def test_click_checkbox_label_with_boolean_field(web_fixture, checkbox_fixture):
    """Clickin on the label of a single checkbox, toggles the checkbox itself."""

    web_fixture.reahl_server.set_app(web_fixture.new_wsgi_app(child_factory=checkbox_fixture.Form.factory()))
    browser = web_fixture.driver_browser
    browser.open('/')

    checkbox = XPath.input_labelled('Subscribe to newsletter?')
    assert not browser.is_checked(checkbox)
    browser.click(XPath.label_with_text('Subscribe to newsletter?'))
    assert browser.is_checked(checkbox)


@with_fixtures(WebFixture, CheckboxFixture)
def test_checkbox_basics_with_multichoice_field(web_fixture, checkbox_fixture):
    """CheckboxInput can also be used to choose many things from a list, in which case it renders many checkboxes."""

    choices = [Choice(1, IntegerField(label='One')),
               Choice(2, IntegerField(label='Two', writable=Action(lambda:False))),
               Choice(3, IntegerField(label='Three')),
               Choice(4, IntegerField(label='Four')),
               ]
    checkbox_fixture.field = MultiChoiceField(choices, label='Make your choice', default=[1])

    web_fixture.reahl_server.set_app(web_fixture.new_wsgi_app(child_factory=checkbox_fixture.Form.factory()))
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.is_element_present(XPath.label_with_text('Make your choice'))

    assert browser.get_xpath_count('//input[@class="custom-control-input"]/following-sibling::label[@class="custom-control-label"]') == 4

    checkbox_one = XPath.input_labelled('One')
    checkbox_two = XPath.input_labelled('Two')
    checkbox_three = XPath.input_labelled('Three')
    checkbox_four = XPath.input_labelled('Four')

    assert browser.is_checked(checkbox_one)
    assert not browser.is_checked(checkbox_two)
    assert not browser.is_element_enabled(checkbox_two)
    # assert browser.is_visible(checkbox_two) #cannot do this as the way bootsrap renders, the actual html input has opacity=0
    assert not browser.is_checked(checkbox_three)
    assert not browser.is_checked(checkbox_four)
    browser.uncheck(checkbox_one)
    browser.check(checkbox_three)
    browser.check(checkbox_four)
    browser.click(XPath.button_labelled('Submit'))
    assert not browser.is_checked(checkbox_one)
    assert browser.is_checked(checkbox_three)
    assert browser.is_checked(checkbox_four)


@uses(web_fixture=WebFixture)
class ChoicesFixture(Fixture):

    def new_form(self):
        return Form(self.web_fixture.view, 'test')

    def new_boolean_field(self):
        field = BooleanField()
        field.bind('field', self)
        return field


@with_fixtures(WebFixture, ChoicesFixture)
def test_choices_layout_applied_to_checkbox(web_fixture, choices_fixture):
    """A ChoicesLayout lays out PrimitiveCheckboxInputs inside a Label containing the Field label, such that they will be stacked."""
    fixture = choices_fixture

    stacked_container = Div(web_fixture.view).use_layout(ChoicesLayout())
    stacked_container.layout.add_choice(PrimitiveCheckboxInput(fixture.form, fixture.boolean_field))

    stacked_container_classes = stacked_container.children[0].get_attribute('class').split(' ')
    assert 'custom-control' in stacked_container_classes
    assert 'custom-checkbox' in stacked_container_classes

    [checkbox_input, label] = stacked_container.children[0].children
    [description_widget] = label.children
    assert label.tag_name == 'label'

    assert checkbox_input.html_representation.input_type == 'checkbox'
    assert description_widget.value == 'field'


@with_fixtures(WebFixture, ChoicesFixture)
def test_checkbox_with_inline_layout(web_fixture, choices_fixture):
    """PrimitiveCheckboxInputs can be rendered inlined with each other by using the ChoicesLayout with the inline=True setting."""
    fixture = choices_fixture

    inlined_container = Div(web_fixture.view).use_layout(ChoicesLayout(inline=True))
    inlined_container.layout.add_choice(PrimitiveCheckboxInput(fixture.form, fixture.boolean_field))

    assert 'custom-control-inline' in inlined_container.children[0].get_attribute('class').split(' ')


@uses(web_fixture=WebFixture)
class DisabledScenarios(Fixture):

    @scenario
    def disabled_input(self):
        self.field = BooleanField(writable=lambda field: False)
        self.expects_disabled_class = True

    @scenario
    def enabled_input(self):
        self.field = BooleanField(writable=lambda field: True)
        self.expects_disabled_class = False


@with_fixtures(WebFixture, DisabledScenarios)
def test_choice_disabled_state(web_fixture, disabled_scenarios):
    """Visible cues are inserted to indicate that inputs are disabled. """
    fixture = disabled_scenarios

    form = Form(web_fixture.view, 'test')
    field = fixture.field
    field.bind('field', fixture)

    container = Div(web_fixture.view).use_layout(ChoicesLayout())
    container.layout.add_choice(PrimitiveCheckboxInput(form, field))

    [checkbox_container] = container.children
    [checkbox_input, label] = checkbox_container.children
    checkbox_container_classes = checkbox_container.get_attribute('class').split(' ')
    if fixture.expects_disabled_class:
        assert 'disabled' in checkbox_container_classes
        assert checkbox_input.html_representation.get_attribute('disabled') == 'disabled'
    else:
        assert 'disabled' not in checkbox_container_classes
        with expected(KeyError):
            checkbox_input.html_representation.get_attribute('disabled')


class RadioButtonFixture(ChoicesFixture):
    def new_field(self):
        choices = [Choice(1, Field(label='One')),
                   Choice(2, Field(label='Two'))]
        field = ChoiceField(choices)
        field.bind('field', self)
        return field


@with_fixtures(RadioButtonFixture)
def test_radio_button_basics(radio_button_fixture):
    """A RadioButtonSelectInput consists of a labelled radio button for each choice, rendered stacked on one another."""
    fixture = radio_button_fixture

    stacked_radio = RadioButtonSelectInput(fixture.form, fixture.field)
    [container] = stacked_radio.children
    [choice1_container, choice2_container] = container.children

    def check_choice_container_details(choice_container, expected_choice_text, expected_button_value):
        choice_container_classes = choice_container.get_attribute('class').split(' ')
        assert 'custom-control' in choice_container_classes
        assert 'custom-radio' in choice_container_classes
        [primitive_radio_button, label] = choice_container.children
        assert label.tag_name == 'label'
        [choice_text_node] = label.children
        assert primitive_radio_button.value == expected_button_value
        assert choice_text_node.value == expected_choice_text

    check_choice_container_details(choice1_container, 'One', '1')
    check_choice_container_details(choice2_container, 'Two', '2')


@with_fixtures(RadioButtonFixture)
def test_radio_button_layout(radio_button_fixture):
    """To make a RadioButtonSelectInput inline, supply a suitable ChoicesLayout"""
    fixture = radio_button_fixture

    radio_input = RadioButtonSelectInput(fixture.form, fixture.field, contents_layout=ChoicesLayout(inline=True))
    assert radio_input.contents_layout.inline

    choice_container = radio_input.children[0].children[0]
    assert 'custom-control-inline' in choice_container.get_attribute('class').split(' ')


@with_fixtures(RadioButtonFixture)
def test_radio_button_label_as_legend(radio_button_fixture):
    """A FormLayout renders a RadioButtonSelectInput in a FieldSet with its label in the Legend."""
    fixture = radio_button_fixture

    form = fixture.form
    form.use_layout(FormLayout())
    inlined_radio = RadioButtonSelectInput(form, fixture.field)
    fixture.form.layout.add_input(inlined_radio)

    field_set = form.children[0]
    assert isinstance(field_set, FieldSet)
    assert 'form-group' in field_set.get_attribute('class').split(' ')

    [label_widget, radio_input_in_form] = form.children[0].children
    assert label_widget.tag_name == 'legend'
    assert 'col-form-label' in label_widget.get_attribute('class')
    assert radio_input_in_form is inlined_radio


@with_fixtures(WebFixture)
def test_button_layouts(web_fixture):
    """A ButtonLayout can be be used on a Button to customise various visual effects."""

    event = Event(label='click me')
    event.bind('event', web_fixture)
    form = Form(web_fixture.view, 'test')
    form.define_event_handler(event)

    # Case: the defaults
    button = Button(form, event).use_layout(ButtonLayout())

    tester = WidgetTester(button)
    [button] = tester.xpath(XPath.button_labelled('click me'))
    assert button.attrib['class'] == 'btn'

    # Case: possible effects
    button = Button(form, event).use_layout(ButtonLayout(style='secondary', size='sm', active=True, wide=True))

    tester = WidgetTester(button)
    [button] = tester.xpath(XPath.button_labelled('click me'))
    assert button.attrib['class'] == 'active btn btn-block btn-secondary btn-sm'


@with_fixtures(WebFixture)
def test_button_layouts_on_anchors(web_fixture):
    """A ButtonLayout can also be used to make an A (anchor) look like a button."""

    anchor = A(web_fixture.view, href=Url('/an/href'), description='link text').use_layout(ButtonLayout())
    tester = WidgetTester(anchor)
    [rendered_anchor] = tester.xpath(XPath.link_with_text('link text'))
    assert rendered_anchor.attrib['class'] == 'btn'


@with_fixtures(WebFixture)
def test_button_layouts_on_disabled_anchors(web_fixture):
    """Disabled A's are marked with a class so Bootstap can style them appropriately."""
    def can_write():
        return False


    anchor = A(web_fixture.view, href=Url('/an/href'), description='link text', write_check=can_write)
    anchor.use_layout(ButtonLayout())

    tester = WidgetTester(anchor)
    [rendered_anchor] = tester.xpath(XPath.link_with_text('link text'))
    assert rendered_anchor.attrib['class'] == 'btn disabled'




# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division
import six
import os
from six.moves import http_cookies 

from webob import Request, Response

from reahl.stubble import stubclass
from reahl.tofu import Fixture, set_up, uses

from reahl.webdev.tools import DriverBrowser
from reahl.webdeclarative.webdeclarative import UserSession, PersistedException, PersistedFile, UserInput

from reahl.domain.systemaccountmodel import LoginSession
from reahl.component.i18n import Catalogue
from reahl.component.py3compat import ascii_as_bytes_or_str
from reahl.component.context import ExecutionContext
from reahl.web.fw import ReahlWSGIApplication, UrlBoundView, UserInterface, Url, Widget
from reahl.web.ui import HTML5Page, Slot, Div
from reahl.web.layout import Layout
from reahl.web.egg import WebConfig


from reahl.dev.fixtures import ReahlSystemFixture
from reahl.webdev.fixtures import WebServerFixture
from reahl.sqlalchemysupport_dev.fixtures import SqlAlchemyFixture
from reahl.domain_dev.fixtures import PartyAccountFixture


_ = Catalogue('reahl-webdev')

        
@stubclass(ReahlWSGIApplication)
class ReahlWSGIApplicationStub(ReahlWSGIApplication):
    def add_reahl_static_files(self):
        static_files = self.config.web.frontend_libraries.packaged_files()
        static_files_no_js = [packaged_file
                              for packaged_file in static_files if not packaged_file.relative_name.endswith('.js')]
        return self.define_static_files('/static', static_files_no_js)


class BasicPageLayout(Layout):
    def __init__(self, slots=['main', 'footer']):
        super(BasicPageLayout, self).__init__()
        self.slots = slots

    def customise_widget(self):
        for slot_name in self.slots:
            slot_div = self.widget.body.add_child(Div(self.view))
            slot_div.append_class('column-%s' % slot_name)
            slot_div.add_child(Slot(self.view, slot_name))


from reahl.web.libraries import Library
class FakeQUnit(Library):
    """
    """
    def __init__(self):
        super(FakeQUnit, self).__init__('fakequnit')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = []

    def footer_only_material(self, rendered_page):
        result = super(FakeQUnit, self).footer_only_material(rendered_page)
        result += '\n<script type="text/javascript">\n'
        result += 'window.QUnit = true;'
        result += '\n</script>\n'
        return result



@uses(reahl_system_fixture=ReahlSystemFixture, 
      party_account_fixture=PartyAccountFixture,
      web_server_fixture=WebServerFixture)
class WebFixture(Fixture):
    """A function-scoped Fixture for use when testing a website via selenium 
    or to test things as if you are inside the server.

    For example, it supplies a View for when you want to instantiate a Widget.
    """
    
    @set_up
    def add_web_config(self):
        self.reahl_system_fixture.config.web = self.webconfig

    @set_up
    def add_request_to_context(self):
        self.reahl_system_fixture.context.request = self.request

    @property
    def context(self):
        """The current :class:`ExecutionContext`."""
        return self.reahl_system_fixture.context

    @property
    def config(self):
        """The current :class:`Configuration`."""
        return self.reahl_system_fixture.config
    
    def new_webconfig(self):
        web = WebConfig()
        web.site_root = UserInterface
        web.static_root = os.path.join(os.getcwd(), 'static')
        web.session_class = UserSession
        web.persisted_exception_class = PersistedException
        web.persisted_file_class = PersistedFile
        web.persisted_userinput_class = UserInput
        web.frontend_libraries.prepend(FakeQUnit())
        return web

    def new_request(self, path=None, url_scheme=None):
        """A :class:`Request` for testing."""
        request = Request.blank(path or '/', charset='utf8')
        request.environ['wsgi.url_scheme'] = url_scheme or 'http'
        if request.scheme == 'http':
            request.environ['SERVER_PORT'] = '8000'
            request.host = 'localhost:8000'
        else:
            request.environ['SERVER_PORT'] = '8363'
            request.host = 'localhost:8363'
        return Request(request.environ, charset='utf8')

    def log_in(self, browser=None, session=None, system_account=None, stay_logged_in=False):
        """Logs the user into the current webapp without having to navigate to a login page."""
        session = session or self.party_account_fixture.session
        browser = browser or self.driver_browser
        login_session = LoginSession.for_session(session)
        login_session.set_as_logged_in(system_account or self.party_account_fixture.system_account, stay_logged_in)
        # quickly create a response so the fw sets the cookies, which we copy and explicitly set on selenium.
        response = Response()
        session.set_session_key(response)
        cookies = http_cookies.BaseCookie(ascii_as_bytes_or_str(', '.join(response.headers.getall('set-cookie'))))
        for name, morsel in cookies.items():
            cookie = {'name':name, 'value':morsel.value}
            cookie.update(dict([(key, value) for key, value in morsel.items() if value]))
            browser.create_cookie(cookie)

    def new_driver_browser(self, driver=None):
        """A :class:`DriverBrowser` instance for controlling the browser."""
        driver = driver or self.web_driver
        return DriverBrowser(driver)

    @property
    def chrome_driver(self):
        """A selenium WebDriver instance for Chromium."""
        return self.web_server_fixture.chrome_driver
    @property
    def firefox_driver(self):
        """A selenium WebDriver instance for Firefox."""
        return self.web_server_fixture.firefox_driver
    @property
    def web_driver(self):
        """The default selenium WebDriver instance."""
        return self.web_server_fixture.web_driver
    @property
    def reahl_server(self):
        """The running Reahl webserver.

        To use it, you will have to call
        :meth:`~reahl.webdev.webserver.ReahlWebServer.set_app` to
        specify the WSGI application to serve.
        """
        return self.web_server_fixture.reahl_server

    def new_wsgi_app(self, site_root=None, enable_js=False, config=None, child_factory=None):
        """Creates a :class:`ReahlWSGIApplication` containing a :class:`UserInterface` with a single View.

        :keyword site_root: an optional supplied :class:`UserInterface` for the webapp 
        :keyword enable_js: If True, JS files will be served. The default is False, to save time.
        :keyword config: A configuration to use.
        :keyword child_factory: A Widget to include in the single-view site (only applicable if no site_root) is given.
        """
        wsgi_app_class = ReahlWSGIApplicationStub
        if enable_js:
            wsgi_app_class = ReahlWSGIApplication
        child_factory = child_factory or Widget.factory()
        config = config or self.config

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(BasicPageLayout())
                view = self.define_view('/', title='Home page')
                view.set_slot('main', child_factory)

        site_root = site_root or MainUI
        config.web.site_root = site_root
        return wsgi_app_class(config)

    @property
    def current_location(self):
        """The current :class:`Url` as displayed in the browser location bar."""
        return Url(self.driver_browser.get_location())

    def new_view(self):
        """An :class:`UrlBoundView` to use when constructing :class:`Widget`\s for testing."""
        current_path = Url(ExecutionContext.get_context().request.url).path
        view = UrlBoundView(self.user_interface, current_path, 'A view')
        return view

    def new_user_interface(self):
        """A :class:`UserInterface` for testing."""
        return UserInterface(None, '/', {}, False, 'test_ui')



# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

from webob import Request, Response
from webob.exc import HTTPNotFound

from reahl.tofu import expected, Fixture
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import stubclass, CallMonitor

from reahl.component.context import ExecutionContext
from reahl.web.fw import Resource, ReahlWSGIApplication, InternalRedirect
from reahl.web.interfaces import UserSessionProtocol
from reahl.dev.fixtures import ReahlSystemFixture
from reahl.web_dev.fixtures import ReahlWSGIApplicationStub
from reahl.webdev.tools import Browser

from reahl.sqlalchemysupport_dev.fixtures import SqlAlchemyFixture
from reahl.web_dev.fixtures import WebFixture


class WSGIFixture(Fixture):
    status = None
    headers = None
    def result_is_valid(self, result):
        return result.startswith('<!DOCTYPE html><html class="no-js">') and result.endswith('</html>')
    def some_headers_are_set(self, headers):
        return dict(headers)['Content-Type'] == 'text/html; charset=utf-8'


@with_fixtures(WebFixture, WSGIFixture)
def test_wsgi_interface(web_fixture, wsgi_fixture):
    """A ReahlWSGIApplication is a WSGI application."""
    fixture = wsgi_fixture
    wsgi_app = web_fixture.wsgi_app

    environ = Request.blank('/', charset='utf8').environ
    def start_response(status, headers):
        fixture.status = status
        fixture.headers = headers

    wsgi_iterator = wsgi_app(environ, start_response)

    result = ''.join([i.decode('utf-8') for i in wsgi_iterator])  # To check that it is iterable and get the value
    assert fixture.result_is_valid(result)
    assert fixture.status == '200 OK'
    assert fixture.some_headers_are_set(fixture.headers)


@with_fixtures(ReahlSystemFixture, WebFixture)
def test_web_session_handling(reahl_system_fixture, web_fixture):
    """The core web framework (this egg) does not implement a notion of session directly.
       It relies on such a notion, but expects an implementation for this to be supplied.

       An implementation should implement the UserSessionProtocol. The implementation
       is thus responsible for associating an instance of itself with the current request.

       The implementation to be used is set in the web.session_class configuration setting.

       The framework ensures that an instance of web.session_class is available during any
       Request. The database is committed before any user code executes, so that any database activity
       done by web.session_class would be committed even if an exception in the user code occurs.
       After user code is executed, methods are called on the web.session_class so that it can set
       its key on the response and save the last_activity_time.
       Finally, another commit is issued to the database so that any database activity during these last
       actions would also be saved.
    """
    stubclass(UserSessionProtocol)
    class UserSessionStub(UserSessionProtocol):
        session = None
        last_activity_time_set = False
        key_is_set = False
        @classmethod
        def for_current_session(cls):
            assert None, 'Not implemented'

        @classmethod
        def get_or_create_session(cls):
            cls.session = cls()
            return cls.session

        @classmethod
        def initialise_web_session_on(cls, context):
            context.session = cls.get_or_create_session()

        def set_session_key(self, response):
            self.key_is_set = True
            self.saved_response = response

        def is_active(self): pass
        def is_secured(self): pass
        def set_as_logged_in(self, party, stay_logged_in): pass
        def log_out(self): pass

        def set_last_activity_time(self):
            self.last_activity_time_set = True

        def get_interface_locale(self):
            return 'en_gb'

    # Setting the implementation in config

    web_fixture.config.web.session_class = UserSessionStub
    with CallMonitor(reahl_system_fixture.system_control.orm_control.commit) as monitor:
        @stubclass(Resource)
        class ResourceStub(object):
            def handle_request(self, request):
                context = ExecutionContext.get_context()
                assert context.session is UserSessionStub.session  # By the time user code executes, the session is set
                assert monitor.times_called == 1  # The database has been committed
                assert not context.session.last_activity_time_set
                assert not UserSessionStub.session.key_is_set
                return Response()

        @stubclass(ReahlWSGIApplication)
        class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
            def resource_for(self, request):
                return ResourceStub()

        browser = Browser(ReahlWSGIApplicationStub2(web_fixture.config))

        # A session is obtained, and the correct params passed to the hook methods
        assert not UserSessionStub.session  # Before the request, the session is not yet set
        assert monitor.times_called == 0  # ... and the database is not yet committed
        browser.open('/')

        assert monitor.times_called == 1  # The database is committed after user code executed
        assert UserSessionStub.session  # The session was set
        assert UserSessionStub.session.key_is_set  # The set_session_key was called
        assert UserSessionStub.session.saved_response.status_int is 200  # The correct response was passed to set_session_key

        assert UserSessionStub.session.last_activity_time_set  # set_last_activity_time was called


@with_fixtures(WebFixture)
def test_handling_HTTPError_exceptions(web_fixture):
    """If an HTTPError exception is raised, it is used as response."""
    @stubclass(ReahlWSGIApplication)
    class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
        def resource_for(self, request):
            raise HTTPNotFound()

    browser = Browser(ReahlWSGIApplicationStub2(web_fixture.config))

    browser.open('/', status=404)


@with_fixtures(WebFixture)
def test_internal_redirects(web_fixture):
    """During request handling, an InternalRedirect exception can be thrown. This is handled by
       restarting the request loop from scratch to handle the same request again, using a freshly
       constructed resource just as though the request was submitted again by the browser
       save for the browser round trip."""

    fixture = web_fixture
    fixture.requests_handled = []
    fixture.handling_resources = []

    @stubclass(Resource)
    class ResourceStub(object):
        def handle_request(self, request):
            fixture.requests_handled.append(request)
            fixture.handling_resources.append(self)
            if hasattr(request, 'internal_redirect'):
                return Response(body='response given after internal redirect')
            raise InternalRedirect(None, None)

    @stubclass(ReahlWSGIApplication)
    class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
        def resource_for(self, request):
            return ResourceStub()

    browser = Browser(ReahlWSGIApplicationStub2(fixture.config))

    browser.open('/')

    assert browser.raw_html == 'response given after internal redirect'
    assert fixture.requests_handled[0] is fixture.requests_handled[1]
    assert fixture.handling_resources[0] is not fixture.handling_resources[1]


@with_fixtures(WebFixture)
def test_handling_uncaught_exceptions(web_fixture):
    """If an uncaught exception is raised, the session is closed properly."""

    @stubclass(ReahlWSGIApplication)
    class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
        def resource_for(self, request):
            raise AssertionError('this an unknown breakage')


    app = ReahlWSGIApplicationStub2(web_fixture.config)
    browser = Browser(app)

    with CallMonitor(app.system_control.finalise_session) as monitor:
        assert monitor.times_called == 0
        with expected(AssertionError):
            browser.open('/')
        assert monitor.times_called == 1



# Copyright 2016, 2017, 2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import os
import threading 

from reahl.tofu import scenario, expected, Fixture, temp_file_with, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.webdev.tools import XPath

from reahl.sqlalchemysupport import Session
from reahl.webdeclarative.webdeclarative import PersistedFile
from reahl.component.modelinterface import exposed, Event, FileField, Action, ValidationConstraint
from reahl.component.exceptions import DomainException
from reahl.web.bootstrap.forms import Form
from reahl.web.bootstrap.files import FileUploadInput, FileUploadPanel, Button, FormLayout, FileInput, FileInputButton

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class FileInputButtonFixture(Fixture):

    def upload_button_indicates_focus(self):
        element = self.web_fixture.driver_browser.find_element(XPath.label_with_text('Choose file(s)'))
        return 'focus' in element.get_attribute('class')

    def new_domain_object(self):
        class DomainObject(object):
            files = []
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label='Attached files')

            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')

        return DomainObject()

    def new_FileUploadForm(self):
        fixture = self
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.add_child(FileInputButton(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))
        return FileUploadForm   


@with_fixtures(WebFixture, FileInputButtonFixture)
def test_file_upload_button(web_fixture, file_input_button_fixture):
    """A FileInputButton lets you upload files using the browser's file choosing mechanism."""

    fixture = file_input_button_fixture


    wsgi_app = web_fixture.new_wsgi_app(child_factory=file_input_button_fixture.FileUploadForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    file_to_upload = temp_file_with('some content')

    browser.type(XPath.input_labelled('Choose file(s)'), file_to_upload.name)

    assert len(fixture.domain_object.files) == 0 
    browser.click(XPath.button_labelled('Submit'))
    assert len(fixture.domain_object.files) == 1 


@with_fixtures(WebFixture, FileInputButtonFixture)
def test_file_upload_button_focus(web_fixture, file_input_button_fixture):
    """If the FileInputButton gets tab focus, it is styled to appear focussed."""

    fixture = file_input_button_fixture


    wsgi_app = web_fixture.new_wsgi_app(child_factory=file_input_button_fixture.FileUploadForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert browser.wait_for_not(fixture.upload_button_indicates_focus) 
    browser.focus_on(XPath.input_labelled('Choose file(s)'))
    assert browser.wait_for(fixture.upload_button_indicates_focus) 


#def test_select_file_dialog_opens(fixture):
#    """Clicking on the FileInputButton opens up the browser choose file dialog."""
#
#    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
#    browser = fixture.driver_browser
#    browser.open('/')
#
#    browser.click(fixture.upload_button_xpath)
#    # We can't dismiss the local file dialog with selenium: one day we would be able to when firefox
#    # selenium is implemented differently


class FileInputFixture(FileInputButtonFixture):
    def new_FileUploadForm(self):
        fixture = self
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.add_child(FileInput(self, fixture.domain_object.fields.files))
        return FileUploadForm   

    message_span_xpath = '//div[contains(@class, "reahl-bootstrapfileinput")]//span[2]'
    def message_displayed_is(self, message):
        return message == self.web_fixture.driver_browser.get_inner_html_for(self.message_span_xpath)

    def message_is_visible(self):
        return self.web_fixture.driver_browser.is_visible(self.message_span_xpath)

    def standard_file_input_is_visible(self):
        input_element = self.web_fixture.driver_browser.find_element(XPath.input_labelled('Choose file(s)'))
        return not input_element.value_of_css_property('width').startswith('0.')


@with_fixtures(WebFixture, FileInputFixture)
def test_file_input_basics(web_fixture, file_input_fixture):
    """A FileInput is a FileInputButton combined with a area where the chosen file name is displayed."""
    fixture = file_input_fixture


    wsgi_app = web_fixture.new_wsgi_app(child_factory=file_input_fixture.FileUploadForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.wait_for(fixture.message_displayed_is, 'No files chosen')
    browser.type(XPath.input_labelled('Choose file(s)'), '/tmp/koos.html')
    browser.wait_for(fixture.message_displayed_is, 'koos.html')

    browser.type(XPath.input_labelled('Choose file(s)'), '/tmp/koos.html\n/tmp/jannie.html')
    browser.wait_for(fixture.message_displayed_is, '2 files chosen')


@with_fixtures(WebFixture, FileInputFixture)
def test_i18n(web_fixture, file_input_fixture):
    """All messages have translations."""
    fixture = file_input_fixture

    wsgi_app = web_fixture.new_wsgi_app(child_factory=file_input_fixture.FileUploadForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/af/')

    browser.wait_for(fixture.message_displayed_is, 'Geen lêers gekies')
    browser.wait_for_element_present(XPath.input_labelled('Kies lêer(s)'))

    browser.type(XPath.input_labelled('Kies lêer(s)'), '/tmp/koos.html\n/tmp/jannie.html')
    browser.wait_for(fixture.message_displayed_is, '2 gekose lêers')


@with_fixtures(WebFixture, FileInputFixture)
def test_file_input_without_js(web_fixture, file_input_fixture):
    """If JS available, we display a bootstrap-styled button, and associated span that look like a pretty version of a standard file input; otherwise we degrade to displaying only the standard file input"""

    fixture = file_input_fixture


    wsgi_app = web_fixture.new_wsgi_app(child_factory=file_input_fixture.FileUploadForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)

    browser = web_fixture.driver_browser
    browser.open('/')

    browser.wait_for(fixture.message_is_visible)
    browser.wait_for_not(fixture.standard_file_input_is_visible)

    browser.switch_styling(javascript=False)

    browser.wait_for_not(fixture.message_is_visible)
    browser.wait_for(fixture.standard_file_input_is_visible)


@uses(web_fixture=WebFixture)
class FileUploadInputFixture(Fixture):

    def file_was_uploaded(self, filename):
        return Session.query(PersistedFile).filter_by(filename=os.path.basename(filename)).count() == 1

    file_to_upload1_name = 'file1.html'
    file_to_upload2_name = 'file2.gif'
    file_to_upload1_content = b'some content'
    file_to_upload2_content = b'some different content'

    def new_file_to_upload1(self):
        return temp_file_with(self.file_to_upload1_content, name=self.file_to_upload1_name, mode='w+b')

    def new_file_to_upload2(self):
        return temp_file_with(self.file_to_upload2_content, name=self.file_to_upload2_name, mode='w+b')

    def new_domain_object(self):
        class DomainObject(object):
            def __init__(self):
                self.throws_exception = False
                self.files = []
                self.submitted_file_info = {}
                self.submitted = False
               
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label='Attached files', required=True)

            @exposed
            def events(self, events):
                events.submit = Event(label='Submit', action=Action(self.submit))
                
            def submit(self):
                if self.throws_exception:
                    raise DomainException()
                for f in self.files:
                    with f.open() as opened_file:
                        contents = opened_file.read()
                    self.submitted_file_info[f.filename] = (contents, f.mime_type)
                self.submitted = True
                
        return DomainObject()

    def new_FileUploadForm(self):
        fixture = self
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.set_attribute('novalidate','novalidate')
                self.use_layout(FormLayout())
                self.layout.add_input(FileUploadInput(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))

        return FileUploadForm

    def new_wsgi_app(self, enable_js=False):
        return self.web_fixture.new_wsgi_app(child_factory=self.FileUploadForm.factory(), enable_js=enable_js)

    def uploaded_file_is_listed(self, filename):
        return self.web_fixture.driver_browser.is_element_present('//ul/li/span[text()="%s"]/../input[@value="Remove"]' % os.path.basename(filename))

    def upload_file_is_queued(self, filename):
        return self.web_fixture.driver_browser.is_element_present('//ul/li/span[text()="%s"]/../input[@value="Cancel"]' % os.path.basename(filename))


class ConstrainedFileUploadInputFixture(FileUploadInputFixture):
    def new_domain_object(self):
        fixture = self
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.files = fixture.file_field

            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')

        return DomainObject()


class PerFileConstrainedFileUploadInputFixture(ConstrainedFileUploadInputFixture):
    @scenario
    def size_constrained(self):
        max_size = 13
        self.file_field = FileField(allow_multiple=True, max_size_bytes=max_size, label='Attached files')
        self.validation_error_message = 'files should be smaller than 13.0bytes'
        self.valid_file = temp_file_with('c'*max_size, name='valid_size.html')
        self.invalid_file = temp_file_with('c'*(max_size+1), name='invalid_size.html')

    @scenario
    def type_constrained(self):
        self.file_field = FileField(allow_multiple=True, accept=['text/*'])
        self.validation_error_message = 'files should be of type text/*'
        self.valid_file = temp_file_with('contents', name='valid.html')
        self.invalid_file = temp_file_with('contents', name='invalid.gif')


class MaxNumberOfFilesFileUploadInputFixture(ConstrainedFileUploadInputFixture):
    def new_file_field(self):
        return FileField(allow_multiple=True, max_files=1)


class ToggleableConstraint(ValidationConstraint):
    def __init__(self, fixture=None):
        super(ToggleableConstraint, self).__init__(error_message='test validation message')
        self.fixture = fixture
        
    def validate_input(self, unparsed_input):
        if self.fixture.make_validation_fail:
            raise self

    def __reduce__(self):
        reduced = super(ToggleableConstraint, self).__reduce__()
        pickle_dict = reduced[2]
        del pickle_dict['fixture']
        return reduced


class ToggleValidationFixture(FileUploadInputFixture):
    make_validation_fail = False
    def new_domain_object(self):
        fixture = self
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label='Attached files')
                fields.files.add_validation_constraint(ToggleableConstraint(fixture))
            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')
                                
        return DomainObject()

    check_script = 'return $(".reahl-nested-form").find(".reload_flag").length > 0'
    def mark_nested_form(self):
        self.web_fixture.driver_browser.execute_script('$(".reahl-nested-form").children().addClass("reload_flag")')
        has_class = self.web_fixture.driver_browser.execute_script(self.check_script)
        assert has_class, 'Something is wrong, could not place flags for checking reloading of form'

    def nested_form_was_reloaded(self):
        has_class = self.web_fixture.driver_browser.execute_script(self.check_script)
        return not has_class  # ie, the UploadPanel has been reloaded


class StubbedFileUploadInputFixture(FileUploadInputFixture):
    run_hook_before = False
    run_hook_after = False
    def new_FileUploadForm(self):
        fixture = self
        class FileUploadInputStub(FileUploadInput):
            def create_html_widget(self):
                return FileUploadPanelStub(self)
            
        class FileUploadPanelStub(FileUploadPanel):
            def upload_file(self):
                if fixture.run_hook_before:
                    fixture.file_upload_hook()
                super(FileUploadPanelStub, self).upload_file()
                if fixture.run_hook_after:
                    fixture.file_upload_hook()

        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.add_child(FileUploadInputStub(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))

        return FileUploadForm


class LargeFileUploadInputFixture(StubbedFileUploadInputFixture):
    def file_upload_hook(self):
        self.simulate_large_file_upload()

    def simulate_large_file_upload(self):
        self.upload_done.wait()

    def simulate_large_file_upload_done(self):
        self.upload_done.set()

    @property
    def web_driver(self):  
        return self.web_fixture.chrome_driver  # These tests only work on chrome

    def new_upload_done(self):
        return threading.Event()


class BrokenFileUploadInputFixture(StubbedFileUploadInputFixture):
    run_hook_after = True
    def file_upload_hook(self):
        raise Exception('simulated exception condition')


class FailingConstraint(ValidationConstraint):
    fail = True
    def validate_input(self, unparsed_input):
        if self.fail:
            raise self


@with_fixtures(WebFixture, FileUploadInputFixture)
def test_file_upload_input_basics(web_fixture, file_upload_input_fixture):
    """A FileUploadInput allows its user to upload multiple files one by one before the Form that
       contains the FileUploadInput is submitted.  When the Form is finally submitted
       the FileField of the FileUploadInput receives all the files uploaded as UploadFile objects.
    """
    fixture = file_upload_input_fixture

    web_fixture.reahl_server.set_app(fixture.wsgi_app)

    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.domain_object.submitted 
    assert not fixture.file_was_uploaded( fixture.file_to_upload1.name ) 
    assert not fixture.file_was_uploaded( fixture.file_to_upload2.name ) 

    # Upload one file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))

    assert not fixture.domain_object.submitted 
    assert fixture.file_was_uploaded( fixture.file_to_upload1.name ) 
    assert not fixture.file_was_uploaded( fixture.file_to_upload2.name ) 

    # Upload a second file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    assert not fixture.domain_object.submitted 
    assert fixture.file_was_uploaded( fixture.file_to_upload1.name ) 
    assert fixture.file_was_uploaded( fixture.file_to_upload2.name ) 

    # Submit the form
    browser.click( XPath.button_labelled('Submit') )
    assert fixture.domain_object.submitted 

    # All uploaded files were submitted
    assert sorted(fixture.domain_object.submitted_file_info.keys()) == sorted([os.path.basename(f.name)
                                                                           for f in [fixture.file_to_upload1, fixture.file_to_upload2]] )

    # Files that were submitted are correct
    file1_content, file1_mime_type = fixture.domain_object.submitted_file_info[os.path.basename(fixture.file_to_upload1.name)]
    assert file1_content == fixture.file_to_upload1_content 
    assert file1_mime_type == 'text/html' 


@with_fixtures(WebFixture, FileUploadInputFixture)
def test_file_upload_input_list_files(web_fixture, file_upload_input_fixture):
    """The FileUploadInput displays a list of files that were uploaded so far, but is cleared 
       once the Form is submitted."""

    fixture = file_upload_input_fixture


    web_fixture.reahl_server.set_app(fixture.wsgi_app)

    browser = web_fixture.driver_browser
    browser.open('/')

    # Upload one file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))

    assert fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    # Upload a second file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    assert fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    # Submit the form:
    # If an exception is raised, the list is NOT cleared
    fixture.domain_object.throws_exception = True
    browser.click( XPath.button_labelled('Submit') )
    assert fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    # Upon successful submit, the list IS cleared
    fixture.domain_object.throws_exception = False
    browser.click( XPath.button_labelled('Submit') )
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 


@with_fixtures(WebFixture, FileUploadInputFixture)
def test_file_upload_input_remove_files(web_fixture, file_upload_input_fixture):
    """A user can remove files that were uploaded but not yet submitted."""
    fixture = file_upload_input_fixture

    web_fixture.reahl_server.set_app(fixture.wsgi_app)

    browser = web_fixture.driver_browser
    browser.open('/')

    # Upload two files
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    # Remove file1
    browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))

    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    # Only the one file is submitted
    browser.click( XPath.button_labelled('Submit') )
    assert list(fixture.domain_object.submitted_file_info.keys()) == [fixture.file_to_upload2_name]


@with_fixtures(WebFixture, FileUploadInputFixture)
def test_file_upload_input_double_uploads(web_fixture, file_upload_input_fixture):
    """The user is prevented from uploading more than one file with the same name.
    """
    fixture = file_upload_input_fixture

    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=False))

    browser = web_fixture.driver_browser
    browser.open('/')

    # Upload two files with the same name
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))

    # Expect an validation error message
    assert browser.is_element_present('//span[text()="uploaded files should all have different names"]') 
    assert fixture.file_was_uploaded(fixture.file_to_upload1.name) 


@with_fixtures(WebFixture, FileUploadInputFixture)
def test_async_upload(web_fixture, file_upload_input_fixture):
    """If JavaScript is enabled, the uploading of files happen in the background via ajax (without reloading the page)
       allowing the user to be busy with the rest of the form. The user does not need to click on the Upload button,
       uploading starts automatically upon choosing a file. The list of uploaded files is appropriately updated.
    """
    fixture = file_upload_input_fixture

    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.file_was_uploaded( fixture.file_to_upload1.name ) 

    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    with browser.no_page_load_expected():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)

    assert browser.get_value(XPath.input_labelled('Choose file(s)')) == ''   # Input is cleared for next file to be input
    assert fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert fixture.file_was_uploaded( fixture.file_to_upload1.name ) 


@with_fixtures(WebFixture, LargeFileUploadInputFixture)
def test_async_in_progress(web_fixture, large_file_upload_input_fixture):
    """While a large file is being uploaded, a progress bar and a Cancel button are displayed. Clicking on the Cancel
       button stops the upload and clears the file name from the list of uploaded files.
    """
    fixture = large_file_upload_input_fixture


    fixture.run_hook_before = True
    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.file_was_uploaded( fixture.file_to_upload1.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 

    with web_fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture

    assert browser.is_element_present('//ul/li/progress') 
    progress = browser.get_attribute('//ul/li/progress', 'value')
    assert progress == '100' 
    browser.click(XPath.button_labelled('Cancel'))

    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert not fixture.file_was_uploaded( fixture.file_to_upload1.name ) 


@with_fixtures(WebFixture, LargeFileUploadInputFixture)
def test_cancelling_queued_upload(web_fixture, large_file_upload_input_fixture):
    """Cancelling an upload that is still queued (upload not started yet) removes the file from the list
       and removed it from the queue of uploads.
    """
    fixture = large_file_upload_input_fixture


    fixture.run_hook_before = True
    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.file_was_uploaded( fixture.file_to_upload1.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert not fixture.file_was_uploaded( fixture.file_to_upload2.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    with web_fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name) # Upload will block, see fixture

    browser.wait_for(fixture.upload_file_is_queued, fixture.file_to_upload1.name)
    browser.wait_for(fixture.upload_file_is_queued, fixture.file_to_upload2.name)

    browser.click(XPath.button_labelled('Cancel', filename=fixture.file_to_upload2_name))

    browser.wait_for_not(fixture.upload_file_is_queued, fixture.file_to_upload2.name)
    fixture.simulate_large_file_upload_done()
    browser.wait_for(fixture.uploaded_file_is_listed, fixture.file_to_upload1.name)

    assert fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert fixture.file_was_uploaded( fixture.file_to_upload1.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 
    assert not fixture.file_was_uploaded( fixture.file_to_upload2.name ) 


@with_fixtures(WebFixture, FileUploadInputFixture)
def test_prevent_duplicate_upload_js(web_fixture, file_upload_input_fixture):
    """The user is prevented from uploading more than one file with the same name on the client side.
    """

    fixture = file_upload_input_fixture


    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = web_fixture.driver_browser

    error_locator = XPath.span_containing('uploaded files should all have different names')
    def error_is_visible():
        return browser.is_visible(error_locator)

    browser.open('/')

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.wait_for_not(error_is_visible)

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.wait_for_not(error_is_visible)

    with web_fixture.reahl_server.paused():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
        assert not fixture.upload_file_is_queued(fixture.file_to_upload1.name) 
        browser.wait_for(error_is_visible)

    browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload2_name))
    browser.wait_for_not(error_is_visible)


@with_fixtures(WebFixture, LargeFileUploadInputFixture)
def test_prevent_form_submit(web_fixture, large_file_upload_input_fixture):
    """The user is prevented from submitting the Form while one or more file uploads are still in progress."""
    fixture = large_file_upload_input_fixture


    fixture.run_hook_after = True
    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = web_fixture.driver_browser
    browser.open('/')

    with web_fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture

    with browser.no_page_load_expected():
        browser.click( XPath.button_labelled('Submit'), wait=False )
        alert = fixture.web_driver.switch_to.alert
        assert alert.text == 'Please try again when all files have finished uploading.' 
        alert.accept()


@with_fixtures(WebFixture, FileUploadInputFixture)
def test_async_remove(web_fixture, file_upload_input_fixture):
    """With javascript enabled, removing of uploaded files take place via ajax."""

    fixture = file_upload_input_fixture


    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = web_fixture.driver_browser
    browser.open('/')

    # Upload two files
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    # Remove file1
    with browser.no_page_load_expected():
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))

    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    # The javascript works on DOM elements that have been generated server-side as well:
    browser.refresh()
    with browser.no_page_load_expected():
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload2_name))

    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    # No files are submitted eventually
    browser.click( XPath.button_labelled('Submit') )
    assert list(fixture.domain_object.submitted_file_info.keys()) == [] 


@with_fixtures(WebFixture, BrokenFileUploadInputFixture)
def test_async_upload_error(web_fixture, broken_file_upload_input_fixture):
    """If an error happens during (ajax) upload, the user is notified."""
    fixture = broken_file_upload_input_fixture

    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    web_fixture.config.reahlsystem.debug = False # So that we don't see the exception output while testing
    browser = web_fixture.driver_browser
    browser.open('/')

    assert not browser.is_element_present(XPath.label_with_text('an error ocurred, please try again later.')) 

    with expected(Exception):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)

    assert browser.wait_for_element_present(XPath.span_containing('an error occurred, please try again later.')) 
    assert not browser.is_element_enabled(XPath.button_labelled('Cancel')) 


@with_fixtures(WebFixture, ToggleValidationFixture)
def test_async_upload_domain_exception(web_fixture, toggle_validation_fixture):
    """When a DomainException happens upon uploading via JavaScript, 
       the form is replaced with a rerendered version from the server."""

    fixture = toggle_validation_fixture

    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = web_fixture.driver_browser
    browser.open('/')

    fixture.make_validation_fail = False
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)

    fixture.make_validation_fail = True
    fixture.mark_nested_form()
    with browser.no_page_load_expected():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    assert fixture.nested_form_was_reloaded()

    # JS Stuff on re-rendered form still work

    # 1: Server-rendered validation message has been cleared
    assert browser.is_visible(XPath.span_containing('test validation message')) 
    fixture.make_validation_fail = False
    with browser.no_page_load_expected():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.wait_for_not(browser.is_visible, XPath.span_containing('test validation message'))

    # 2: The remove button still happens via ajax
    with browser.no_page_load_expected():
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload2_name))


@with_fixtures(WebFixture, LargeFileUploadInputFixture)
def test_queueing_async_uploads(web_fixture, large_file_upload_input_fixture):
    """Asynchronous uploads do not happen concurrently, they are queued one after another.
    """
    fixture = large_file_upload_input_fixture


    fixture.run_hook_after = True
    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.file_was_uploaded(fixture.file_to_upload1.name) 
    assert not fixture.uploaded_file_is_listed(fixture.file_to_upload1.name) 

    with web_fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name) # Upload will block, see fixture

    progress1 = browser.get_attribute('//ul/li[1]/progress', 'value')
    assert progress1 == '100'
    progress2 = browser.get_attribute('//ul/li[2]/progress', 'value')
    assert progress2 == '0'

    fixture.simulate_large_file_upload_done()
    browser.wait_for( fixture.uploaded_file_is_listed, fixture.file_to_upload2.name )

    assert fixture.uploaded_file_is_listed(fixture.file_to_upload1.name) 
    assert fixture.uploaded_file_is_listed(fixture.file_to_upload2.name) 
    assert fixture.file_was_uploaded(fixture.file_to_upload1.name) 
    assert fixture.file_was_uploaded(fixture.file_to_upload2.name) 


@with_fixtures(WebFixture, PerFileConstrainedFileUploadInputFixture)
def test_async_validation(web_fixture, per_file_constrained_file_upload_input_fixture):
    """Validations are checked in JavaScript before uploading.
    """
    # Only tested for the FileUploadInput, as it uses the FileInput
    # in its own implementation, in a NestedForm, and has to pass on the
    # filesize constraint all the way. This way, we test all of that.
    fixture = per_file_constrained_file_upload_input_fixture

    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.uploaded_file_is_listed( fixture.valid_file.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.invalid_file.name ) 

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.invalid_file.name)
    assert not fixture.uploaded_file_is_listed( fixture.invalid_file.name ) 
    assert browser.is_element_present(XPath.span_containing(fixture.validation_error_message)) 

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.valid_file.name)
    assert fixture.uploaded_file_is_listed( fixture.valid_file.name ) 


@with_fixtures(WebFixture, MaxNumberOfFilesFileUploadInputFixture)
def test_async_number_files_validation(web_fixture, max_number_of_files_file_upload_input_fixture):
    """A Field set to only allow a maximum number of files is checked for validity before uploading in JS.
    """
    # Only tested for the FileUploadInput, as it uses the FileInput
    # in its own implementation, in a NestedForm, and has to pass on the
    # filesize constraint all the way. This way, we test all of that.
    fixture = max_number_of_files_file_upload_input_fixture

    web_fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    assert fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) 
    # Corner case: max are uploaded, but you've not asked to add to them yet:
    assert browser.wait_for_not(browser.is_visible, XPath.span_containing('a maximum of 1 files may be uploaded')) 

    # Normal case: max are uploaded, and you're asking to upload another:
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    assert not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) 
    assert browser.wait_for(browser.is_visible, XPath.span_containing('a maximum of 1 files may be uploaded')) 

    browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))
    assert browser.wait_for_not(browser.is_visible, XPath.span_containing('a maximum of 1 files may be uploaded')) 



# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division
import six


from reahl.tofu import Fixture, scenario, expected, uses, NoException
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import EmptyStub, stubclass

from reahl.web.ui import HTMLElement, PrimitiveInput, Form, CheckboxInput, TextInput, Label, ButtonInput,\
                          PasswordInput, TextArea, SelectInput, RadioButtonSelectInput, CheckboxSelectInput

from reahl.component.modelinterface import Field, EmailField, BooleanField, Event, Allowed, exposed, \
    Action, Choice, ChoiceGroup, ChoiceField, IntegerField, MultiChoiceField, DateField
from reahl.component.exceptions import IsInstance, ProgrammerError
from reahl.webdev.tools import WidgetTester
from reahl.webdev.tools import XPath

from reahl.web_dev.fixtures import WebFixture
from reahl.component_dev.test_field import FieldFixture


@uses(web_fixture=WebFixture, field_fixture=FieldFixture)
class SimpleInputFixture(Fixture):

    def new_field(self):  # not a delegated property, because some scenarios override this by setting self.field
        return self.field_fixture.new_field()

    def new_model_object(self): # not a delegated property, because some classes override it
        return self.field_fixture.new_model_object()

    def new_form(self):
        return Form(self.web_fixture.view, 'test')

    def new_event(self, label='click me'):
        event = Event(label=label)
        event.bind('aname', None)
        self.form.define_event_handler(event)
        return event


class SimpleInputFixture2(SimpleInputFixture):
    def new_field(self, label='the label'):
        return Field(label=label)
    def new_model_object(self):
        fixture = self
        class ModelObject(object):
            def handle_event(self):
                pass
            @exposed
            def events(self, events):
                events.an_event = Event(label='click me', action=Action(self.handle_event))
            @exposed
            def fields(self, fields):
                fields.an_attribute = fixture.field
        return ModelObject()

    def new_Form(self, input_widget_class=None):
        fixture = self
        input_widget_class = input_widget_class or CheckboxSelectInput
        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                checkbox = self.add_child(input_widget_class(self, fixture.model_object.fields.an_attribute))
                self.define_event_handler(fixture.model_object.events.an_event)
                self.add_child(ButtonInput(self, fixture.model_object.events.an_event))
                fixture.checkbox = checkbox
        return MyForm


class InputStateFixture(SimpleInputFixture):
    def new_field(self):
        field = EmailField(default='default value')
        field.bind('an_attribute', self.model_object)
        return field

    def new_input(self):
        @stubclass(PrimitiveInput)
        class InputStub(PrimitiveInput):
            def create_html_widget(self):
                return HTMLElement(self.view, 'input')

        return InputStub(self.form, self.field)

    @scenario
    def defaulted_input_no_value_on_domain(self):
        del self.model_object.an_attribute
        self.input.prepare_input()

        self.expected_state = 'defaulted'
        self.expected_value = 'default value'

    @scenario
    def defaulted_input_value_on_domain(self):
        self.input.prepare_input()
        self.model_object.an_attribute = 'something'

        self.expected_state = 'defaulted'
        self.expected_value = 'something'

    @scenario
    def validly_entered(self):
        self.input.enter_value('someone@home.org')
        self.input.prepare_input()

        self.expected_state = 'validly_entered'
        self.expected_value = 'someone@home.org'

    @scenario
    def empty_input_non_required_field(self):
        self.input.enter_value('')
        self.input.prepare_input()

        self.expected_state = 'validly_entered'
        self.expected_value = ''

    @scenario
    def empty_input_required_field(self):
        self.field.make_required('')
        self.input.enter_value('')
        self.input.prepare_input()

        self.expected_state = 'invalidly_entered'
        self.expected_value = ''

    @scenario
    def invalidly_entered(self):
        self.input.enter_value('not an email address')
        self.input.prepare_input()

        self.expected_state = 'invalidly_entered'
        self.expected_value = 'not an email address'


@with_fixtures(WebFixture, InputStateFixture)
def test_the_states_of_an_input(web_fixture, input_state_fixture):
    """An Input can be in one of three states: defaulted, invalidly_entered or validly_entered. Depending
       on the state of the input, it will have a different value when rendering.

       The states:
         - defaulted: the user has not entered anything
           value:     the value on the model object which is bound to its Field if that object
                      has such an an_attribute, else the default value of the Field.
         - invalidly_entered: the user entered a value which failed validation
           value:             the invalid value which the user entered
         - validly_entered: the user entered a valid value
           value:           the value which the user entered even if the underlying
                            model object has a corresponding value
    """
    fixture = input_state_fixture

    assert fixture.input.get_input_status() == fixture.expected_state
    assert fixture.input.value == fixture.expected_value


class InputScenarios(SimpleInputFixture):
    @scenario
    def text_input(self):
        self.widget = TextInput(self.form, self.field)
        self.expected_html = '<input name="an_attribute" form="test" type="text" value="field value" class="reahl-textinput">'
        self.field_controls_visibility = True

    @scenario
    def text_input_placeholder_default(self):
        self.widget = TextInput(self.form, self.field, placeholder=True)
        self.expected_html = '<input name="an_attribute" aria-label="the label" form="test" placeholder="the label" type="text" value="field value" class="reahl-textinput">'
        self.field_controls_visibility = True

    @scenario
    def text_input_placeholder_specified(self):
        self.widget = TextInput(self.form, self.field, placeholder="some text")
        self.expected_html = '<input name="an_attribute" aria-label="some text" form="test" placeholder="some text" type="text" value="field value" class="reahl-textinput">'
        self.field_controls_visibility = True

    @scenario
    def input_label(self):
        html_input = TextInput(self.form, self.field)
        self.widget = Label(self.web_fixture.view, for_input=html_input)
        self.expected_html = '<label for="%s">the label</label>' % html_input.css_id
        self.field_controls_visibility = True

    @scenario
    def input_label_with_text(self):
        html_input = TextInput(self.form, self.field)
        self.widget = Label(self.web_fixture.view, for_input=html_input, text='some text')
        self.expected_html = '<label for="%s">some text</label>' % html_input.css_id
        self.field_controls_visibility = True

    @scenario
    def button_input(self):
        self.widget = self.form.add_child(ButtonInput(self.form, self.event))
        self.expected_html = '<input name="event.aname?" form="test" type="submit" value="click me">'
        self.field_controls_visibility = False

    @scenario
    def password(self):
        self.widget = self.form.add_child(PasswordInput(self.form, self.field))
        self.expected_html = '<input name="an_attribute" form="test" type="password">'
        self.field_controls_visibility = True

    def setup_checkbox_scenario(self, boolean_value):
        self.model_object.an_attribute = boolean_value

        self.field = BooleanField(required=True, label='my text', required_message='$label is needed here')
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(CheckboxInput(self.form, self.field))
        self.field_controls_visibility = True

    @scenario
    def checkbox_true(self):
        """A checkbox needs a 'checked' an_attribute if its field is True. It also renders ONLY the
            validation message for its required validation_constraint, not for all constraints"""
        self.setup_checkbox_scenario(True)
        self.expected_html = '<input name="an_attribute" checked="checked" data-msg-required="my text is needed here" form="test" required="*" type="checkbox">'

    @scenario
    def checkbox_false(self):
        self.setup_checkbox_scenario(False)
        self.expected_html = '<input name="an_attribute" data-msg-required="my text is needed here" form="test" required="*" type="checkbox">'

    @scenario
    def text_area_input(self):
        self.widget = self.form.add_child(TextArea(self.form, self.field, rows=30, columns=20))
        self.expected_html = '<textarea name="an_attribute" cols="20" rows="30">field value</textarea>'
        self.field_controls_visibility = True

    @scenario
    def select_input(self):
        self.model_object.an_attribute = 2

        choices = [Choice(False, BooleanField(label='None')),
                   ChoiceGroup('grouped', [
                       Choice(1, IntegerField(label='One')),
                       Choice(2, IntegerField(label='Two'))])
                  ]
        self.field = ChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(SelectInput(self.form, self.field))
        group = '<optgroup label="grouped"><option value="1">One</option><option selected="selected" value="2">Two</option></optgroup>'
        option = '<option value="off">None</option>'
        self.expected_html = r'<select name="an_attribute" form="test">%s%s</select>' % (option, group)
        self.field_controls_visibility = True

    @scenario
    def select_input_multi(self):
        self.model_object.an_attribute = [2]

        choices = [Choice(1, IntegerField(label='One')),
                   Choice(2, IntegerField(label='Two'))]
        self.field = MultiChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(SelectInput(self.form, self.field))
        options = '<option value="1">One</option><option selected="selected" value="2">Two</option>'
        self.expected_html = '<select name="an_attribute" form="test" multiple="multiple">%s</select>' % (options)
        self.field_controls_visibility = True

    @scenario
    def checkbox_input_multi(self):
        selected_checkboxes = [2, 3]
        self.model_object.an_attribute = selected_checkboxes

        choices = [Choice(1, IntegerField(label='One')),
                   Choice(2, IntegerField(label='Two')),
                   Choice(3, IntegerField(label='Three'))
                  ]
        self.field = MultiChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(CheckboxSelectInput(self.form, self.field))
        not_checked_option = r'<label><input name="an_attribute" form="test" type="checkbox" value="1">One</label>'
        checked_options = r'<label><input name="an_attribute" checked="checked" form="test" type="checkbox" value="2">Two</label>'
        checked_options += r'<label><input name="an_attribute" checked="checked" form="test" type="checkbox" value="3">Three</label>'
        options = not_checked_option + checked_options
        self.expected_html = r'<div class="reahl-checkbox-input">%s</div>' % (options)
        self.field_controls_visibility = True

    @scenario
    def radio_button(self):
        self.model_object.an_attribute = 2

        choices = [ Choice(1, IntegerField(label='One')),
                    Choice(2, IntegerField(label='Two'))]
        self.field = ChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(RadioButtonSelectInput(self.form, self.field))

        radio_button = '<label>'\
                       '<input name="an_attribute"%s '\
                              'data-msg-pattern="an_attribute should be one of the following: 1|2" '\
                              '%s'\
                              'form="test" pattern="(1|2)" '\
                              'title="an_attribute should be one of the following: 1|2" '\
                              'type="radio" value="%s">%s'\
                       '</label>'

        outer_div = '<div class="reahl-radio-button-input">%s</div>'
        buttons = (radio_button % ('', '', '1', 'One')) +\
                  (radio_button % (' checked="checked"', '', '2', 'Two'))
        self.expected_html = outer_div % buttons
        self.field_controls_visibility = True


@with_fixtures(InputScenarios)
def test_basic_rendering(input_scenarios):
    """What the rendered html for a number of simple inputs look like."""

    fixture = input_scenarios

    tester = WidgetTester(fixture.widget)
    actual = tester.render_html()
    assert actual == fixture.expected_html


@with_fixtures(InputScenarios)
def test_rendering_when_not_allowed(input_scenarios):
    """When not allowed to see the Widget, it is not rendered."""
    fixture = input_scenarios

    tester = WidgetTester(fixture.widget)

    fixture.field.access_rights.readable = Allowed(False)
    fixture.field.access_rights.writable = Allowed(False)
    actual = tester.render_html()
    if fixture.field_controls_visibility:
        assert actual == ''
    else:
        assert actual == fixture.expected_html


@with_fixtures(FieldFixture, SimpleInputFixture)
def test_input_wrapped_widgets(field_fixture, simple_input_fixture):
    """An Input is an empty Widget; its contents are supplied by overriding its
       .create_html_widget() method. Several methods for setting HTML-things, like
       css id are delegated to this Widget which represents the Input in HTML.
    """
    fixture = simple_input_fixture

    class MyInput(PrimitiveInput):
        def create_html_widget(self):
            return HTMLElement(self.view, 'x')

    test_input = MyInput(fixture.form, field_fixture.field)
    tester = WidgetTester(test_input)

    rendered = tester.render_html()
    assert rendered == '<x>'

    test_input.set_id('myid')
    test_input.set_title('mytitle')
    test_input.add_to_attribute('list-attribute', ['one', 'two'])
    test_input.set_attribute('an-attribute', 'a value')

    rendered = tester.render_html()
    assert rendered == '<x id="myid" an-attribute="a value" list-attribute="one two" title="mytitle">'


@with_fixtures(SimpleInputFixture)
def test_wrong_args_to_input(simple_input_fixture):
    """Passing the wrong arguments upon constructing an Input results in an error."""

    fixture = simple_input_fixture

    with expected(IsInstance):
        PrimitiveInput(fixture.form, EmptyStub())

    with expected(IsInstance):
        PrimitiveInput(EmptyStub(), Field())


@with_fixtures(WebFixture, SimpleInputFixture2)
def test_marshalling_of_checkbox_input(web_fixture, checkbox_fixture):
    """When a form is submitted, the value of a checkbox is derived from
       whether the checkbox is included in the submission or not."""

    fixture = checkbox_fixture
    fixture.field = BooleanField(label='my text')

    model_object = fixture.model_object
    wsgi_app = web_fixture.new_wsgi_app(child_factory=fixture.new_Form(input_widget_class=CheckboxInput).factory('myform'))
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    # Case: checkbox is submitted with form (ie checked)
    web_fixture.driver_browser.check("//input[@type='checkbox']")
    web_fixture.driver_browser.click("//input[@value='click me']")

    assert model_object.an_attribute
    assert fixture.checkbox.value == 'on'

    # Case: checkbox is not submitted with form (ie unchecked)
    web_fixture.driver_browser.uncheck("//input[@type='checkbox']")
    web_fixture.driver_browser.click("//input[@value='click me']")

    assert not model_object.an_attribute
    assert fixture.checkbox.value == 'off'


@with_fixtures(WebFixture, SimpleInputFixture2)
def test_marshalling_of_checkbox_select_input(web_fixture, checkbox_fixture):
    """CheckboxSelectInput is used to choose many things from a list."""

    fixture = checkbox_fixture
    choices = [Choice(1, IntegerField(label='One')),
               Choice(2, IntegerField(label='Two')),
               Choice(3, IntegerField(label='Three'))]
    fixture.field = MultiChoiceField(choices)

    model_object = fixture.model_object
    model_object.an_attribute = [1]
    wsgi_app = web_fixture.new_wsgi_app(child_factory=fixture.new_Form(input_widget_class=CheckboxSelectInput).factory('myform'))
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    assert web_fixture.driver_browser.is_checked(XPath.input_labelled('One'))
    assert not web_fixture.driver_browser.is_checked(XPath.input_labelled('Two'))
    assert not web_fixture.driver_browser.is_checked(XPath.input_labelled('Three'))

    # Case: checkbox is submitted with form (ie checked)
    web_fixture.driver_browser.uncheck(XPath.input_labelled('One'))
    web_fixture.driver_browser.check(XPath.input_labelled('Two'))
    web_fixture.driver_browser.check(XPath.input_labelled('Three'))
    web_fixture.driver_browser.click(XPath.button_labelled('click me'))

    assert model_object.an_attribute == [2, 3]
    assert fixture.checkbox.value == ['2', '3']
    assert not web_fixture.driver_browser.is_checked(XPath.input_labelled('One'))
    assert web_fixture.driver_browser.is_checked(XPath.input_labelled('Two'))
    assert web_fixture.driver_browser.is_checked(XPath.input_labelled('Three'))


class RadioButtonInputFieldScenarios(Fixture):
    @scenario
    def choice_field(self):
        self.default_domain_value = 1
        self.selected_domain_value = 2
        self.field_class = ChoiceField

    @scenario
    def multi_choice_field(self):
        self.default_domain_value = [1]
        self.selected_domain_value = [2]
        self.field_class = MultiChoiceField


@with_fixtures(WebFixture, SimpleInputFixture2, RadioButtonInputFieldScenarios)
def test_marshalling_of_radio_button_select_input(web_fixture, input_fixture, field_scenario):
    """When a form is submitted, the value of the selected radio button is persisted."""

    fixture = input_fixture
    choices = [Choice(1, IntegerField(label='One')),
               Choice(2, IntegerField(label='Two'))]
    fixture.field = field_scenario.field_class(choices)

    model_object = fixture.model_object
    model_object.an_attribute = field_scenario.default_domain_value

    wsgi_app = web_fixture.new_wsgi_app(child_factory=fixture.new_Form(input_widget_class=RadioButtonSelectInput).factory('myform'))
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    radio_one = XPath.input_labelled('One')
    radio_two = XPath.input_labelled('Two')
    # One is pre selected
    assert web_fixture.driver_browser.is_checked(radio_one)
    assert not web_fixture.driver_browser.is_checked(radio_two)

    # click on Two
    web_fixture.driver_browser.click(radio_two)
    assert not web_fixture.driver_browser.is_checked(radio_one)

    web_fixture.driver_browser.click(XPath.button_labelled('click me'))

    assert model_object.an_attribute == field_scenario.selected_domain_value


@with_fixtures(WebFixture)
def test_checkbox_input_restricted_to_use_with_boolean(web_fixture):
    """CheckboxInput is for toggling a true or false answer."""
    model_object = web_fixture
    form = Form(web_fixture.view, 'test')

    # case: with BooleanField
    boolean_field = BooleanField(label='Boolean')
    boolean_field.bind('a_choice', model_object)
    with expected(NoException):
        CheckboxInput(form, boolean_field)

    # case: with disallowed field
    not_a_boolean_field = IntegerField(label='Other')
    not_a_boolean_field.bind('not_a_boolean', model_object)

    with expected(IsInstance):
        CheckboxInput(form, not_a_boolean_field)


class ManyChoicesScenarios(Fixture):
    @scenario
    def checkbox_select_input(self):
        self.input_type = CheckboxSelectInput
        self.xpath_function_to_choice = XPath.input_labelled

    #@scenario
    def select_input(self):
        self.input_type = SelectInput
        self.xpath_function_to_choice = XPath.option_with_text


@with_fixtures(WebFixture, SimpleInputFixture2, ManyChoicesScenarios)
def test_choices_disabled(web_fixture, checkbox_fixture, choice_type_scenario):
    """A choice that is not writable renders as disabled."""

    fixture = checkbox_fixture
    choices = [Choice(1, IntegerField(label='One')),
               Choice(2, IntegerField(label='Two', writable=Allowed(False)))]
    fixture.field = MultiChoiceField(choices)

    model_object = fixture.model_object
    model_object.an_attribute = [1]
    wsgi_app = web_fixture.new_wsgi_app(child_factory=fixture.new_Form(input_widget_class=choice_type_scenario.input_type).factory('myform'))
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    assert web_fixture.driver_browser.is_element_enabled(choice_type_scenario.xpath_function_to_choice('One'))
    assert not web_fixture.driver_browser.is_element_enabled(choice_type_scenario.xpath_function_to_choice('Two'))


class FuzzyTextInputFixture(SimpleInputFixture2):
    def new_field(self, label='the label'):
        return DateField(label=label)


@with_fixtures(WebFixture, FuzzyTextInputFixture)
def test_fuzzy(web_fixture, fuzzy_text_input_fixture):
    """A TextInput can be created as fuzzy=True. Doing this results in the possibly imprecise
       input that was typed by the user to be interpreted server-side and changed to the
       more exact representation in the client browser.  This happens upon the input losing focus."""
    fixture = fuzzy_text_input_fixture

    model_object = fixture.model_object
    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(TextInput(self, model_object.fields.an_attribute, fuzzy=True))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(ButtonInput(self, model_object.events.an_event))

    wsgi_app = web_fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.type(XPath.input_named('an_attribute'), '20 November 2012')
    browser.press_tab()
    browser.wait_for(browser.is_element_value, XPath.input_named('an_attribute'), '20 Nov 2012')




# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

import pkg_resources

from reahl.tofu import temp_dir, Fixture, set_up, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import easter_egg

from reahl.webdev.tools import Browser

from reahl.web.fw import ReahlWSGIApplication, UserInterface
from reahl.web.ui import HTML5Page
from reahl.web.libraries import Library

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class LibraryFixture(Fixture):

    def new_egg_dir(self):
        egg_dir = temp_dir()
        package_dir = egg_dir.sub_dir('static_files')
        init_file = package_dir.file_with('__init__.py', '')
        js_file = package_dir.file_with('somefile.js', 'contents - js')
        css_file = package_dir.file_with('somefile.css', 'contents - css')
        return egg_dir

    def new_MyLibrary(self):
        easter_egg.clear()
        pkg_resources.working_set.add(easter_egg)
        easter_egg.location = self.egg_dir.name

        class MyLibrary(Library):
            def __init__(self):
                super(MyLibrary, self).__init__('mylib')
                self.files = ['somefile.js', 'somefile.css']
                self.shipped_in_directory = '/static_files'
                self.egg_name = easter_egg.project_name

        return MyLibrary

    @set_up
    def configure_site_root(self):
        self.web_fixture.webconfig.site_root = self.MainUI

    def new_MainUI(self):
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_view('/', title='Home page')
        return MainUI


@with_fixtures(WebFixture, LibraryFixture)
def test_configuring_libraries(web_fixture, library_fixture):
    """Reahl can be configured to expose frontend libraries (libraries of js and css files)."""

    web = web_fixture.webconfig
    assert 'mylib' not in web.frontend_libraries
    web.frontend_libraries.add(library_fixture.MyLibrary())
    assert 'mylib' in web.frontend_libraries


@with_fixtures(WebFixture, LibraryFixture)
def test_library_files(web_fixture, library_fixture):
    """The files part of configured frontend libraries are (a) added to /static and also (b) included on any page."""

    config = web_fixture.config
    config.web.frontend_libraries.clear()
    config.web.frontend_libraries.add(library_fixture.MyLibrary())

    browser = Browser(ReahlWSGIApplication(config))

    browser.open('/static/somefile.js')
    assert browser.raw_html == 'contents - js'

    browser.open('/static/somefile.css')
    assert browser.raw_html == 'contents - css'

    browser.open('/')
    script_added = browser.get_html_for('//script[@src]')
    assert script_added == '<script type="text/javascript" src="/static/somefile.js"></script>'

    link_added = browser.get_html_for('//link')
    assert link_added == '<link rel="stylesheet" href="/static/somefile.css" type="text/css">'


@with_fixtures(WebFixture)
def test_standard_reahl_files(web_fixture):
    """The framework includes certain frontent frameworks by default."""

    config = web_fixture.config
    wsgi_app = ReahlWSGIApplication(config)
    browser = Browser(wsgi_app)

    browser.open('/static/html5shiv-printshiv-3.7.3.js')
    assert browser.last_response.content_length > 0

    browser.open('/static/IE9.js')
    assert browser.last_response.content_length > 0

    browser.open('/static/reahl.validate.js')
    assert browser.last_response.content_length > 0

    browser.open('/static/reahl.css')
    assert browser.last_response.content_length > 0

    browser.open('/static/runningon.png')
    assert browser.last_response.content_length > 0

    browser.open('/static/reahl.runningonbadge.css')
    assert browser.last_response.content_length > 0



# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import six

from reahl.tofu import Fixture, expected, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import EmptyStub, stubclass

from reahl.webdev.tools import WidgetTester, Browser

from reahl.component.exceptions import IncorrectArgumentError, IsInstance
from reahl.web.fw import UrlBoundView, UserInterface, Widget
from reahl.web.ui import Div, P, Slot


from reahl.web_dev.fixtures import WebFixture



@with_fixtures(WebFixture)
def test_basic_widget(web_fixture):
    """Basic widgets are created by adding children widgets to them, and the result is rendered in HTML"""

    class MyMessage(Div):
        def __init__(self, view):
            super(MyMessage, self).__init__(view)
            self.add_child(P(view, text='Hello World!'))
            self.add_children([P(view, text='a'), P(view, text='b')])

    fixture = web_fixture

    message = MyMessage(fixture.view)

    widget_tester = WidgetTester(message)
    actual = widget_tester.render_html()

    assert actual == '<div><p>Hello World!</p><p>a</p><p>b</p></div>'


@with_fixtures(WebFixture)
def test_visibility(web_fixture):
    """Widgets are rendered only if their .visible property is True."""

    class MyMessage(Widget):
        is_visible = True
        def __init__(self, view):
            super(MyMessage, self).__init__(view)
            self.add_child(P(view, text='你好世界!'))

        @property
        def visible(self):
            return self.is_visible

    fixture = web_fixture

    message = MyMessage(fixture.view)

    widget_tester = WidgetTester(message)

    # Case: when visible
    message.is_visible = True
    actual = widget_tester.render_html()
    assert actual == '<p>你好世界!</p>'

    # Case: when not visible
    message.is_visible = False
    actual = widget_tester.render_html()
    assert actual == ''


@with_fixtures(WebFixture)
def test_widget_factories_and_args(web_fixture):
    """Widgets can be created from factories which allow you to supply widget-specific args
       and/or kwargs before a view is available."""

    class WidgetWithArgs(Widget):
        def __init__(self, view, arg, kwarg=None):
            super(WidgetWithArgs, self).__init__(view)
            self.saved_arg = arg
            self.saved_kwarg = kwarg

    fixture = web_fixture

    factory = WidgetWithArgs.factory('a', kwarg='b')

    widget = factory.create(fixture.view)

    assert widget.saved_arg == 'a'
    assert widget.saved_kwarg == 'b'


def test_widget_factories_error():
    """Supplying arguments to .factory that do not match those of the Widget's __init__ is reported to the programmer.."""

    class WidgetWithArgs(Widget):
        def __init__(self, view, arg, kwarg=None):
            super(WidgetWithArgs, self).__init__(view)
            self.saved_arg = arg
            self.saved_kwarg = kwarg

    expected_message = 'An attempt was made to create a WidgetFactory for %r with arguments that do not match what is expected for %r' % (WidgetWithArgs, WidgetWithArgs)
    with expected(IncorrectArgumentError, test=expected_message):
        WidgetWithArgs.factory('a', 'b', 'c')


def test_widget_construct_error():
    """Passing anything other than a View as a Widget's view argument on construction results in an error."""

    with expected(IsInstance):
        Widget(EmptyStub())


@with_fixtures(WebFixture)
def test_widget_adding_error(web_fixture):
    """Passing anything other than other Widgets to .add_child or add_children results in an error."""
    fixture = web_fixture

    widget = Widget(fixture.view)

    with expected(IsInstance):
        widget.add_child(EmptyStub())

    with expected(IsInstance):
        widget.add_children([Widget(fixture.view), EmptyStub()])


@with_fixtures(WebFixture)
def test_basic_working_of_slots(web_fixture):
    """Slots are special Widgets that can be added to the page. The contents of a
       Slot are then supplied (differently) by different Views."""

    class MyPage(Widget):
        def __init__(self, view):
            super(MyPage, self).__init__(view)
            self.add_child(Slot(view, 'slot1'))
            self.add_child(Slot(view, 'slot2'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(MyPage)

            home = self.define_view('/', title='Home')
            home.set_slot('slot1', P.factory(text='a'))
            home.set_slot('slot2', P.factory(text='b'))

            other = self.define_view('/other', title='Other')
            other.set_slot('slot1', P.factory(text='other'))

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')
    [slot1_p, slot2_p] = browser.lxml_html.xpath('//p')
    assert slot1_p.text == 'a'
    assert slot2_p.text == 'b'

    browser.open('/other')
    [slot1_p] = browser.lxml_html.xpath('//p')
    assert slot1_p.text == 'other'


@with_fixtures(WebFixture)
def test_defaults_for_slots(web_fixture):
    """A Widget can have defaults for its slots."""

    class MyPage(Widget):
        def __init__(self, view):
            super(MyPage, self).__init__(view)
            self.add_child(Slot(view, 'slot3'))
            self.add_default_slot('slot3', P.factory(text='default'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(MyPage)
            self.define_view('/', title='Home')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')
    [slot3_p] = browser.lxml_html.xpath('//p')
    assert slot3_p.text == 'default'


@with_fixtures(WebFixture)
def test_activating_javascript(web_fixture):
    """The JavaScript snippets of all Widgets are collected in a jQuery.ready() function by"""
    """ an automatically inserted Widget in the slot named reahl_footer.  Duplicate snippets are removed."""
    @stubclass(Widget)
    class WidgetWithJavaScript(Widget):
        def __init__(self, view, fake_js):
            super(WidgetWithJavaScript, self).__init__(view)
            self.fake_js = fake_js

        def get_js(self, context=None):
           return [self.fake_js]

    class MyPage(Widget):
        def __init__(self, view):
            super(MyPage, self).__init__(view)
            self.add_child(WidgetWithJavaScript(view, 'js1'))
            self.add_child(WidgetWithJavaScript(view, 'js2'))
            self.add_child(WidgetWithJavaScript(view, 'js1'))#intended duplicate
            self.add_child(Slot(view, 'reahl_footer'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(MyPage)
            self.define_view('/', title='Home')

    fixture = web_fixture
    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')
    rendered_js = [i.text for i in browser.lxml_html.xpath('//script[@id="reahl-jqueryready"]')][0]
    assert rendered_js == '\njQuery(document).ready(function($){\n$(\'body\').addClass(\'enhanced\');\njs1\njs2\n\n});\n'

    number_of_duplicates = rendered_js.count('js1') - 1
    assert number_of_duplicates == 0
